/*{{{  label.c*/
/*****************************************************************
PROBLEM:  Marking CASE DEFAULT blocks -- 'yyparse' must distinguish
'CASE DEFAULT' blocks from other CASE blocks (viz. in former case call
'begin_default_block()', but currently it doesn't.)

Also, we must buffer-up the default block for o/p later (at end
of CASE constr!!)

BUG
---
-- Doesn't handle labelled 'do-term-action-stmt' correctly.  It's
transformed to a DO-term "CONTINUE" ** followed** by the action stmt.
That's because the DO-term is o/p by 'set_part_and_label()',
which is called *before* any o/p for the stmt itself!
	To correct will need this to return a flag, and then
check whether DO-term o/p needed at end of every stmt!

-- "GOTO n" doesn't give correct label if it's a 'DO label'.
*****************************************************************/

/*{{{  COMMENT -- Minor bugs in label handling*/
/*----------------------------------------------------------------------
Minor bugs in label handling
----------------------------
(1) Useless labels / symbols are left lying around:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This happens if:
	(we're in a CONTAINS but not in a proc therein)  &&
	(stmt_part == DONT_TEST_PART).

Stmts to which this applies are:
	ELSEIF, ELSE, ENDIF, etc.
	END INTERFACE, END TYPE, etc.
	syntax errors.

(N.B. For most stmts, with an actual 'stmt_part', 'set_part_and_label()'
clears up unwanted symbols, labels, etc, introduced by stmts in illegal
context!)

(2) Incorrect association of label with unit:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Since a stmt label is installed by 'analyse_stmt()' right at the start
of lexing the stmt, before any parsing is done, then it's associated with
the wrong unit if the stmt itself unexpectedly changes the unit, viz:
-- (i) if its a unit header stmt not preceded by an end-unit stmt;
-- (ii) if it's an end-unit stmt for an *outer* unit (not the current one);

Examples are:
(1)
---
   fn f (a,b)
   ...
   goto 10
   ...
   contains
     subr g(...)
       ...
10 end function f	! "missing END stmt for subr 'G'".
			!    ...suggesting that this stmt assoc with fn 'F'.
			!    However, also get this conflicting errmsg about
			!    the 'GOTO 10' stmt in 'F':
			! "label '10' not set (referenced in line 15)"
(2)
---
    type a
      real aa
      real bb
10  x = y	! "illegal stmt in typedef -- missing 'END TYPE' stmt assumed"
		!     ... suggesting this stmt part of prog unit.
		!     N.B. If this were a decl, X & Y would be installed as 
		!     if they're in typedef!!
    goto 10
    end		! "label '10' not set"

(3)
---
10  a = b + 10
    ...
10  fn f (a,b,c)	! now, get:
			!   "label '10' already set (at line 1)"
			!   "missing END stmt for main program"
goto 10
end			!   "label '10' not set"


----------------------------------------------------------------------*/

/*}}}*/

/** sort out 'n_where_errs' ***/


/*{{{  global decls*/

#include <stdio.h>
#define STDIO_H
#include "label.h"
#include "lex.h"
#include "output.h"
#include "strmacs.h"
#include "tokens.h"
#include "typedef.h"
#include "unit.h"


/*--------------------------------------*/
/* definition of globals in 'label.h'	*/
/*--------------------------------------*/
bool  just_written_adapt_finish = FALSE;

int   stmt_label = 0, stmt_newlabel = 0;

Label  *stmt_lp;
                          
int  n_where_errs = 0;	/****** RE-INITIALISE THIS ??? ****/


/*--------------------------------------------------------------*/
/* 'Blockrec' -- a record for every block, stored in order of	*/
/* textual occurrence.						*/
/*--------------------------------------------------------------*/
Blockrec  *first_blrecp;	/* pointer to first block record!! */
int  n_blocks;			/* total number of blocks in 'Blockrec' */


/*--------------------------------------------------------------*/
/* 'Blockstack' -- a stack of the blocks currently 'in scope'.	*/
/*--------------------------------------------------------------*/
Blockstack  first_blstk;	/* the 1st 'block stack' record */
				/* -- static for efficiency     */

Blockstack  *top_blstkp = &first_blstk;
				/* ptr to top record of block stack.	*/
				/* N.B. This *must* be initialised to	*/
				/* point to the static 'first_blstk'.	*/
int  *current_blockp;		/* pointer to 'current_blockno' */
int  current_blockno = 0;	/* pointed to by 'current_blockp' */
int  n_blocks_in_scope;		/* number of blocks 'in scope' (i.e. in Blockstack') */

/*}}}*/

/*{{{  set_part_and_label (stmt_part, token)*/
set_part_and_label (stmt_part, token)
/*---------------------------------*/
	int  stmt_part,		/* may be 0 (== DONT_TEST_PART !) */
	     token;		/* non-zero for stmt types referenced herein */
/*----------------------------------------------------------------
	This is called by 'yyparse' after parsing each statement,
to perform the following actions:

(i) If (stmt_part != 0), check that the statement is appropriate
to the current 'program part', which is adjusted if necessary.

(ii) If the statement is labelled, install details of the 'type' of
statement in 'lp->type', so that the validity of references to the
label can be checked.  If the label has already been referenced,
check that these references are valid.  If not, change its 'newlabel'
(which was assigned by 'lookup_or_install_label()') if it's a FORMAT
stmt or cannot be referenced (and is therefore redundant).

MINOR BUG: If an illegal stmt appears in a type def, an errmsg states:
	"missing 'END TYPE' assumed"
... but that only takes effect *after* the illegal stmt has been
processed (because this fn is normally called then).
----------------------------------------------------------------*/
{
start:
	if (stmt_part != DONT_TEST_PART) {
		/*{{{  check 'stmt_part', and if necessary adjust 'unitp->attrs'*/
		/*----------------------------------------------------------------
		Check that 'stmt_part' is appropriate, and if necessary adjust the
		UNIT_PART flag in 'this_unitp->attrs'.
		
		N.B. This bit isn't done if (stmt_part == DONT_TEST_PART), which is
		used for statements that shouldn't affect the 'prog_part',  e.g. ones
		that terminate a block, like ELSEIF, ELSE, ENDIF, or a unit, like END,
		END FUNCTION, ENDTYPE, or a CONTAINS stmt.
			If they're incorrectly placed (viz. not within the
		appropriate construct or unit) there'll be an errmsg to that effect,
		but an errant ENDIF stmt in the decls part, for example, doesn't
		indicate the start of the exec part -- according to the language
		definition, only 'action stmts' and 'executable construct header'
		stmts do that!
			Also, if we're currently not in a proc unit, we don't want
		any special errmsgs to be given or actions to be performed for these
		stmts -- their normal errmsgs (e.g. "ENDIF not in an IF construct") suffice.
			Label types ('lp->type') for such stmts are determined
		individually if a specific token is given for the stmt type
		(e.g. TK_END_IF), or are set to LB_ILLEGAL_TARGET otherwise.
		----------------------------------------------------------------*/
		int  unit_attrs = this_unitp->attrs;
		
		if (stmt_part & unit_attrs)	/* ok! */
			;
		
		else if (stmt_part > unit_attrs) {
			/*{{{  increment the 'unit part'.*/
			/*--------------------------------------------------------------*/
			/* Increment the program part.  If the statement might belong	*/
			/* to any of several program parts, increment to the lowest one	*/
			/* possible.							*/
			/*--------------------------------------------------------------*/
			do {
				increment_unit_part (unit_attrs);
				if (unit_attrs & EXEC_PART) {
					process_decls ();
					if (unit_attrs & INTERFACE_BODY_LEVEL)
						errmsg ("executable statements not allowed in interface body");
			
					else if (unit_attrs & EXTERNAL_LEVEL)
						write_decls ();
				}
			} while ( !(stmt_part & unit_attrs));
			
			this_unitp->attrs = unit_attrs;
			/*}}}*/
		}
		
		else if (unit_attrs & UNIT_PART) {
			/*{{{  error -- stmt belongs to a previous program part.*/
			/*--------------------------------------------------------------*/
			/* Error!  Statement belongs to a previous program part!	*/
			/*--------------------------------------------------------------*/
			switch (stmt_part) {
			case  USE_PART :
				errmsg ("\'USE\' statements must precede all other statements in a program unit");
				break;
			
			case  IMPNONE_PART :
				/*--------------------------*/
				/* 'unit part' >= IMPL_PART */
				/*--------------------------*/
				errmsg ("\'IMPLICIT NONE\' must be the first statement in a program unit apart from \'USE\' statements");
				break;
			
			case  IMPL_PART :
				/*--------------------------*/
				/* 'unit part' >= DECL_PART */
				/*--------------------------*/
				errmsg ("\'IMPLICIT\' statements must precede all other statements in a program unit apart from \'USE\' and \'PARAMETER\' statements");
				break;
			
			case  DECL_PART :
			case  IMPL_OR_DECL_PART :
				/*----------------------------------------*/
				/* 'unit part' == EXEC_PART or WHERE_PART */
				/*----------------------------------------*/
				errmsg ("declarations must precede all executable statements");
				break;
			
			default :
				/*------------------------------------------------------*/
				/* Stmt is executable  &&  'unit part' == WHERE_PART.	*/
				/* Report an error if the stmt is not an assignment.	*/
				/* To avoid the possibility of this errmsg being	*/
				/* repeated until the end of the procedure, after 3	*/
				/* such errors we assume that 'END WHERE' is missing	*/
				/* and terminate the 'WHERE' block.			*/
				/*------------------------------------------------------*/
				{
					Block *bp = lookup_current_block ();
			
					if (token == TK_ASSIGNMENT)
						;
					else if (++n_where_errs < 3)	
						errmsg ("only assignments are allowed in a WHERE block (which begins on line %d)",
							bp->block_lineno);
					else {
						errmsg ("only assignments are allowed in a WHERE block (which begins on line %d);  missing \'ENDWHERE\' is assumed",
							bp->block_lineno);
						end_current_block ();
					}
					break;
				}
			}
			/*}}}*/
		}
		
		else if (unit_attrs & NOT_IN_PROC_UNIT) {
			/*{{{  in a context where only a subset of stmts allowed*/
			/*--------------------------------------------------------------*/
			/* In a context where only a small subset (possibly empty) of	*/
			/* F90 stmts is allowed.					*/
			/*--------------------------------------------------------------*/
			switch (unit_attrs & UNIT_LEVEL) {
			case  EXTERNAL_LEVEL :
			
				/*{{{  assume that the stmt begins the MAIN program unit*/
				/*----------------------------------------------*/
				/* assume that the stmt begins the MAIN unit.	*/
				/*----------------------------------------------*/
				process_hdr_stmt (TK_PROGRAM,  NULL_Sym,  0,  NULL_str,  NULL_Sym);
				set_unit_part (this_unitp->attrs,  stmt_part);
				break;
				/*}}}*/
			
			case  INTERNAL_LEVEL :
			case  MODULE_LEVEL :
			
				/*{{{  in a CONTAINS part, but not in a proc*/
				/*--------------------------------------------------------------*/
				/* In 'CONTAINS' part, but not in proc.  Here, we undo any	*/
				/* actions resulting from the stmt, e.g. the installation of	*/
				/* labels, blocks, and 'symtab' and 'imptab' entries.		*/
				/*--------------------------------------------------------------*/
				errmsg ("illegal statement in \'CONTAINS\' part: only procedures are allowed");
				
				if (this_unitp->first_sp)	clear_local_symtab ();
				
				if (this_unitp->first_lp)	clear_labels ();
				
				if (token == TK_IMPLICIT)	init_imptab (this_unitp);
				
				if (n_blocks_in_scope)     	clear_blocks ();
				
				return;		/* don't want to process stmt label!! */
				/*}}}*/
			
			case  INTERFACE_BODY_LEVEL :
			case  INTERFACE_MOD_PROC_LEVEL :
			
				/*{{{  in an INTERFACE block, but not in a proc*/
				/*---------------------------------------*/
				/* in 'INTERFACE' block, but not in proc */
				/*---------------------------------------*/
				if (token == TK_MODULE_PROC) {
					set_unit_level (this_unitp->attrs, INTERFACE_MOD_PROC_LEVEL);
					break;
				}
				errmsg ("illegal statement in INTERFACE block: only procedures and \'MODULE PROCEDURE\' statements are allowed");
				fatal_error ();
				break;
				/*}}}*/
				
			case  TYPEDEF_PRIV_SEQ_LEVEL :
				/*------------------------------------------------------*/
				/* At the start of a TYPE definition -- expecting a	*/
				/* PRIVATE, SEQUENCE or type declaration stmt.		*/
				/* N.B. This code *isn't* invoked for a SEQUENCE stmt.	*/
				/*------------------------------------------------------*/
				if (token == TK_TYPE_DECL)
					set_unit_level (this_unitp->attrs, TYPEDEF_COMPNT_LEVEL);
			
				/* now fall through to next case ... */
				
			case  TYPEDEF_COMPNT_LEVEL :
				/*------------------------------------------------------*/
				/* In a TYPE definition -- expecting a type declaration	*/
				/* stmt.						*/
				/*------------------------------------------------------*/
				if (token == TK_ACCESS  || token == TK_TYPE_DECL)
					break;		/* any necessary errmsgs already given */
				
				errmsg ("illegal statement in derived-type definition - missing \'END TYPE\' assumed");
				close_current_unit ();
				goto start;	/* try again! */
			}
			/*}}}*/
		}
		/*}}}*/

/*** in a TYPE def & WHERE constr, what HPF directives are allowed?? ***/

	}
	if (stmt_lp) {
		/*{{{  assign label 'type'*/
		/*--------------------------------------------------------------*/
		/* Assign label 'type'  (e.g. LB_FORMAT, LB_TARGET, etc.)	*/
		/*--------------------------------------------------------------*/
		switch (token) {
		
		/*----------------------------------------------*/
		/*  Format statement  (type = LB_FORMAT)	*/
		/*----------------------------------------------*/
		case  TK_FORMAT:
			stmt_lp->type = LB_FORMAT;
			break;
		
		/*----------------------------------------------*/
		/*  Illegal targets  (type = LB_ILLEGAL_TARGET)	*/
		/*----------------------------------------------*/
		/*{{{  ELSE, ELSEIF, CASE, DEFAULT*/
		case TK_ELSE :
			stmt_lp->type = LB_ILLEGAL_ELSE;
			break;
			
		case TK_ELSE_IF :
			stmt_lp->type = LB_ILLEGAL_ELSE_IF;
			break;
		
		case TK_CASE :
		case TK_DEFAULT :
			stmt_lp->type = LB_ILLEGAL_CASE;
			break;
			
		
		/*}}}*/
		
		/*--------------------------------------------------------------*/
		/* Valid targets but unsuitable as DO-terms  (type = LB_TARGET)	*/
		/*--------------------------------------------------------------*/
		/*{{{  Construct stmts, various kinds of GOTOs, STOP, RETURN and END stmts. */
		
		/*--------------------------------------------------------------*/
		/* Construct stmts (excl. ENDDO!) are unsuitable as DO-term	*/
		/* stmts.							*/
		/* N.B. We exclude the in-construct stmts ELSEIF, ELSE and	*/
		/* CASE from this list, as they are singled-out as illegal	*/
		/* targets (type = LB_ILLEGAL_TARGET).  Also, we exclude ENDIF,	*/
		/* which has its own special type (LB_END_IF).			*/
		/*--------------------------------------------------------------*/
		
		case TK_DO :
		case TK_DO_WHILE :
		case TK_DO_FOREVER :
		case TK_IF_CONSTR :
		case TK_SELECT_CASE:
		case TK_WHERE_CONSTR :
		
		case TK_ELSE_WHERE :
		
		case TK_END_SELECT :
		case TK_END_WHERE :
		
		/*--------------------------------------------------------------*/
		/* Unconditional 'gotos', of which there are 3 kinds, are also	*/
		/* illegal as DO-term stmts.  They are the first 3 cases below.	*/
		/*								*/
		/* N.B. 'computed_goto' is okay, as it may not result in a	*/
		/* branch being taken (it's like a primitive case construct).	*/
		/*--------------------------------------------------------------*/
		
		case TK_GO_TO :
		case TK_ASSIGNED_GO_TO :
		case TK_ARITH_IF :	/* this is yet another kind of GO_TO! */
		
		/*--------------------------------------------------------------*/
		/* Other valid branch targets that are unsuitable as DO-terms.	*/
		/*--------------------------------------------------------------*/
		case TK_CYCLE :
		case TK_EXIT :
		
		case TK_RETURN :
		case TK_STOP :
		case TK_END :
		
			stmt_lp->type = LB_TARGET | current_blockno;
			break;
		
		
		
		/*}}}*/
		
		/*--------------------------------------------------------------*/
		/* ENDIF is a special case. It's a valid target, but if		*/
		/* branched to from the same 'blockno' as the ENDIF stmt, this	*/
		/* corresponds to jumping to the ENDIF from outside its		*/
		/* IF-block, which must be flagged as OBSOLESCENT usage!!	*/
		/*--------------------------------------------------------------*/
		case TK_END_IF :
			stmt_lp->type = LB_END_IF | current_blockno;
			break;
		
		/*--------------------------------------------------------------*/
		/* ENDDO should be handled by 'case default', but it's a valid	*/
		/* branch target whose 'stmt_part' == DONT_TEST_PART  rather	*/
		/* than EXEC_PART! All other such cases are handled explicitly	*/
		/* above.							*/
		/*--------------------------------------------------------------*/
		case TK_END_DO :
			stmt_lp->type = LB_TARGET_OR_DO_TERM | current_blockno;
			break;
			
		/*----------------------------------------------*/
		/*  Otherwise...				*/
		/*----------------------------------------------*/
		default :
			if (stmt_part & EXEC_PART)
				stmt_lp->type = LB_TARGET_OR_DO_TERM | current_blockno;
			else
				stmt_lp->type = LB_ILLEGAL_TARGET;
			break;
		}
		
		/*}}}*/
		if (stmt_lp->last_refp) {
			/*{{{  check label refs (N.B. DO-terms are processed here!)*/
			/*--------------------------------------------------------------*/
			/* Check label references in reverse order of their textual	*/
			/* occurrence.  If the statement terminates a DO-loop, it is	*/
			/* processed here.						*/
			/*								*/
			/* N.B. We go through the label refs in reverse order to	*/
			/* facilitate the handling of DO-loops that share a single DO-	*/
			/* termination stmt, as these must be processed innermost-first,*/
			/* and the innermost loop corresponds to the last DO-statement.	*/
			/*--------------------------------------------------------------*/
			
			Labelref *refp;
			int  n_do_ref = 0;
			
			for (refp = stmt_lp->last_refp;  refp;  refp = refp->prev) {
			
				switch (refp->type & LB_TYPE_MASK) {
				case LB_DO_REF :
					/*{{{  is the stmt a valid DO-term?*/
					/*--------------------------------------------------------------*/
					/* If this is the first reference to the DO-term statement by	*/
					/* a DO-statement, check its validity as a DO-termination and	*/
					/* give an error message if it's invalid.			*/
					/*								*/
					/* (However, even if it's an invalid DO-term, we still call	*/
					/* 'end_labelled_do_loop()' to continue with the subsequent	*/
					/* checking and block-handling -- i.e. we terminate the DO-loop	*/
					/* as if it were a valid DO-termination statement).		*/
					/*--------------------------------------------------------------*/
					
					if ((++n_do_ref == 1)  &&  !(stmt_lp->type & LB_TARGET_OR_DO_TERM)) {
						/*----------------------*/
						/* not a valid DO-term! */
						/*----------------------*/
						Labelref  *lrefp;
						int  total_do_refs;
					
						/*--------------------------------------------------------------*/
						/* count the total number of DO-stmts referring to this DO-term */
						/*--------------------------------------------------------------*/
						total_do_refs = 1;
						for (lrefp = refp->prev;  lrefp;  lrefp = lrefp->prev) {
							if (lrefp->type & LB_DO_REF)
								total_do_refs++;
						}
					
						/*----------------------------------------------*/
						/* give an errmsg (listing all the DO-refs!!)	*/
						/*----------------------------------------------*/
						begin_errmsg ("illegal DO-loop termination ");
						if (total_do_refs == 1)
							end_errmsg ("(of the \'DO %d\' loop starting on line %d)",
								stmt_lp->label, refp->lineno);
						else {
							int  i;
							
							cont_errmsg ("(of the \'DO %d\' loops starting on lines %d",
								stmt_lp->label, refp->lineno);
					
							i = 1;
							for (lrefp = refp->prev;  lrefp;  lrefp = lrefp->prev) {
								if ( !(lrefp->type & LB_DO_REF))
									continue;
								else if (++i == total_do_refs)
									end_errmsg (" and %d)", lrefp->lineno);
								else
									cont_errmsg (", %d", lrefp->lineno);
							}
						}
					}
					if ( ! (token == TK_CONTINUE  ||  token == TK_END_DO))
						not_supported ("termination of a DO loop at a statement other than ENDDO or CONTINUE");
					/*}}}*/
					end_labelled_do_loop (refp, token, n_do_ref);
					break;
			
				case LB_GO_TO_REF :
					check_goto_ref (refp, stmt_lp);
					break;
			
				case LB_FORMAT_REF :
					if (token != TK_FORMAT)
						errmsg ("label %d on line %d must refer to a FORMAT statement",
							stmt_lp->label, refp->lineno);
					break;
			
				case LB_ASSIGN_REF :
					check_assign_ref (refp, stmt_lp);
					break;
			
				}
			}
		/*}}}*/
		}
		else {
			/*{{{  */
			/*--------------------------------------------------------------
				'newlabel' was assigned as soon as the statement label was
			seen, prior to parsing the statement.  (This is done so that if a
			reference to another new label occurs in the statement, its new
			number will be assigned in sequence).  The newlabel is assigned
			on the assumption that the statement is a branch target.  This is
			inappropriate in two cases:
			
			(i)	for a FORMAT statement (since its newlabel comes from a
				different sequence, counting down from 99999 rather than up
				from 1);
			
			(ii)	for a statement that is an illegal target (in which
				case it's better not to output a new label at all!).
			
			In these cases the new labels are re-assigned here.  Note that the
			statement does not reference any labels in either case, so there's
			no risk of interfering with the sequence of new label numbers.
			----------------------------------------------------------------*/
			
			switch (stmt_lp->type & LB_TYPE_MASK) {
			case LB_FORMAT :
				stmt_lp->newlabel = stmt_newlabel = change_newlabel_to_fmt ();
				break;
				
			case LB_ILLEGAL_TARGET :
				stmt_lp->newlabel = stmt_newlabel = remove_newlabel ();
									/* set it to 0! */
				break;
			}
			/*}}}*/
		}
	}
	else if (token == TK_FORMAT)
		errmsg ("unlabelled FORMAT statement");

	/*{{{  reset 'just_written_adapt_finish' if necessary*/
	if (token == TK_RETURN  ||  token == TK_STOP)
		just_written_adapt_finish = TRUE;
	
	else if (token != TK_END  ||  stmt_lp)
		just_written_adapt_finish = FALSE;
	
	/* ... otherwise this is an unlabelled 'END' stmt, in which case
	we want to preserve the value that 'just_written_adapt_finish' had
	after the immediately preceding stmt. */
	/*}}}*/
}
			/*}}}*/

/*{{{  <<< label handling functions >>>*/
/*------------------------------*/
/* Label handling functions	*/
/*------------------------------*/
/*{{{  Label *lookup_or_install_label (int label,  bool set_label) -- ok*/
Label *lookup_or_install_label (label,  set_label)
/*----------------------------------------------*/
	int   label;
	bool  set_label;	/* TRUE if label set, not referenced */
/*----------------------------------------------------------------
	Lookup 'label' in the label list and install it if it
isn't there, returning a pointer to its entry.
	This function is called whenever a label is encountered, viz:

-- to set a stmt label (by 'get_f77_stmt_line()', 'analyse_stmt()'
	and 'missing_end_for_procs_upto()");

-- or for a label reference (by 'install_label_ref ()').

The former case is distinguished by (set_label == TRUE).
The function assumes that the label is valid.
	If the label is newly installed, a candidate 'newlabel'
is also set, assuming that the label *doesn't* belong to (i.e.
label or refer to) a FORMAT statement.  If it does, the mistake
is corrected by 'set_part_and_label()' or 'install_newlabel ()'
respectively before another label is processed.
	If (set_label == TRUE), an errmsg is given if 'lp->lineno'
is already set (i.e. the label has already been used).  Otherwise
'lp->lineno' is installed, and 'stmt_lp', 'stmt_label' and
'stmt_newlabel' are all set.
	'lp->type' is determined and installed later.
----------------------------------------------------------------*/
{
	Label *lp;

	for (lp = this_unitp->first_lp;  lp;  lp = lp->next)
		if (lp->label == label)
			break;		/* label found! */

	if (! lp) {
		/*{{{  create an entry for 'label', and set candidate 'newlabel'*/
		/*--------------------------------------------------------------*/
		/* Create an entry for 'label'.					*/
		/* N.B. lp->type, lp->lineno, lp->refp  are set later.		*/
		/*--------------------------------------------------------------*/
		lp = (Label *) ecalloc (sizeof (Label));	/* zero fields! */
		lp->label = label;
		if (label != LB_UNLABELLED_DO)
			/*------------------------------------------------------*/
			/* Provided that 'label' is 'real' (i.e. not just a	*/
			/* token for an unlabelled DO-loop) install a 'newlabel'*/
			/* assuming it's not a FORMAT stmt or reference -- see	*/
			/* the explanation above!				*/
			/*------------------------------------------------------*/
			lp->newlabel = get_newlabel (LB_TARGET);
		
		/*----------------------------------------------*/
		/* Add it to the head of the label list.	*/
		/*----------------------------------------------*/
		lp->next = this_unitp->first_lp;
		this_unitp->first_lp = lp;
		/*}}}*/
	}

	if (set_label) {
		/*{{{  set 'lp->lineno', 'stmt_lp / label / newlabel'*/
		if (lp->lineno)
			errmsg ("label %d has already been used (at line %d)",
				label, lp->lineno);
		else {
			if (use_current_lineno)
				/*----------------------------------------------*/
				/* Fixed form input.  Label installed as line	*/
				/* is being input, by 'get_f77_stmt_line ()'.	*/
				/*----------------------------------------------*/
				lp->lineno = lineno;
			else {
				/*----------------------------------------------*/
				/* Free form input.  Label installed after line	*/
				/* has been input, by 'analyse_stmt ()'.	*/
				/*----------------------------------------------*/
				int  i;
		
				for (i = 0;  start_tk > stmt_endp [i];  i++)
					;
				lp->lineno = stmt_lineno [i];
			}
			stmt_lp       = lp;
			stmt_label    = label;
			stmt_newlabel = lp->newlabel;
		}
		/*}}}*/
	}
	
	return  lp;
}



/*}}}*/
/*{{{  Label* install_label_ref (int label, type) -- OK*/
Label *install_label_ref (label, type)
/*----------------------------------*/
	int  label, type;
/*----------------------------------------------------------------
	Checks that the 'label' is valid, and if so, creates a
label reference record (a 'Labelref' record, comprising the block
and line number of the reference), and attaches it to the label
record (which is created if it doesn't already exist).
	'type', the type of the label reference, is one of
LB_DO_REF / _FORMAT_REF / _GO_TO_REF / _ASSIGN_REF.
All label refs are bound to be legal by definition!  However,
not all labelled stmts are valid targets for label refs!
	If the label has already been defined (i.e. there has
already been a statement with that label), then the validity of
the reference is checked; otherwise, the validity of all forward
references is checked when the labelled statement is encountered.
	The function returns a pointer to the label record
(not the label *reference* record!).
----------------------------------------------------------------*/
{
	Label *lp;
	Labelref *refp;

	/*{{{  take appropriate action if the label is invalid*/
	/*--------------------------------------------------------------*/
	/*	Take appropriate action if the label is invalid!	*/
	/*								*/
	/* N.B.1:  The grammar rules are such that negative numbers are	*/
	/* not recognised as labels -- labels are guaranteed to be >= 0.*/
	/* Hence, certain negative values are used to flag special cases*/
	/* (e.g. as a label placeholder for unlabelled DO-loops).	*/
	/*								*/
	/* N.B.2:  Perhaps (?) these label validity tests are		*/
	/* insufficient - perhaps we should also test that the label	*/
	/* has no more than 5 digits(?)  (e.g. should an error be	*/
	/* flagged for a label of 012345 ???).				*/
	/*--------------------------------------------------------------*/
	{
		if (label == 0 || label >= 100000) {
			/*--------------------------------------*/
			/* invalid label -- give errmsg...	*/
			/*--------------------------------------*/
			if (label == 0)
				errmsg ("invalid label - must be non-zero");
			else
				errmsg ("invalid label \'%d\' - must not have more than 5 digits",
					 label);
	
			/*----------------------------------------------*/
			/* take appropriate corrective action...	*/
			/*----------------------------------------------*/
			switch (type) {
			case LB_DO_REF :
				/*--------------------------------------*/
				/* treat it as an unlabelled DO...	*/
				/*--------------------------------------*/
				label = LB_UNLABELLED_DO;
				break;
			default :
				return  NULL;
			}
		}
	}
	/*}}}*/
	
	/*{{{  lookup or install the label*/
	
	lp = lookup_or_install_label (label, REF_LABEL);
	
	/*--------------------------------------------------------------*/
	/* If the label refers to a FORMAT statement and this is its	*/
	/* first appearance, then re-assign the 'newlabel' (which was	*/
	/* assigned by 'lookup_or_install_label()' on the assumption	*/
	/* that it's a branch target!).					*/
	/*--------------------------------------------------------------*/
	if ((type == LB_FORMAT_REF)  &&  ! (lp->last_refp || lp->lineno))
		lp->newlabel = change_newlabel_to_fmt ();
	/*}}}*/
	
	/*{{{  create and install the 'Labelref' record*/
	
	refp = (Labelref *) emalloc (sizeof (Labelref));
	
	refp->type = type | current_blockno;
		/*------------------------------------------------------*/
		/* ...N.B. The 'current_blockno' is only useful for	*/
		/* GOTO and DO references, but for simplicity we might	*/
		/* as well always record it!				*/
		/*------------------------------------------------------*/
	
	refp->lineno = lineno;
	
	if (type == LB_DO_REF)
		lp->newlabel = refp->do_newlabel = get_newlabel (LB_DO_TERM);
		/*------------------------------------------------------*/
		/* N.B.:						*/
		/* -- 'refp->do_blockno' is installed by 'yyparse'	*/
		/* after this function returns (and after it has called	*/
		/* 'begin_block (TK_DO)'!).				*/
		/* -- We change 'lp->newlabel' to that of the		*/
		/* transformed Do-term so that subsequent refs (e.g.	*/
		/* "GOTO n") have the correct newlabel.  Note that in	*/
		/* this case there should have been no previous refs	*/
		/* unless there were previous 'LB_DO_REF's (in which	*/
		/* case this is a shared DO-term stmt!).		*/
		/*------------------------------------------------------*/
	
		
	/*--------------------------------------------------------------*/
	/* Attach 'refp' to the label record, 'lp', at the end of the	*/
	/* linked list of label references:				*/
	/*--------------------------------------------------------------*/
	
	refp->prev    =  lp->last_refp;		/* may be NULL! */
	lp->last_refp = refp;
	/*}}}*/

	if (lp->lineno)	{
		/*{{{  labelled stmt already encountered - check the ref*/
		switch (type) {
		case LB_FORMAT_REF :
			if ( !(lp->type & LB_FORMAT) )
				errmsg ("statement with label %d (on line %d) is not a FORMAT statement",
					lp->label, lp->lineno);
			break;
				
		case LB_GO_TO_REF :
			check_goto_ref (refp, lp);
			break;
				
		case LB_ASSIGN_REF :
			check_assign_ref (refp, lp);
			break;
				
		case LB_DO_REF :
			/*------------------------------------------------------*/
			/* This shouldn't happen -- it means that a DO-stmt	*/
			/* appears *after* the DO-term, or is its *own* DO-term!*/
			/*							*/
			/* N.B. In this case, the DO-block is still created	*/
			/* because the action for 'do_stmt' (in 'yyparse')	*/
			/* calls 'begin_block (TK_DO)'.  However, it won't be	*/
			/* terminated (as the terminal label has already been	*/
			/* used), and we'll get an errmsg to this effect at the	*/
			/* end of the unit.  That's ok - it emphasises the	*/
			/* error!!						*/
			/*------------------------------------------------------*/
			if ((lp->lineno) < lineno)
				errmsg ("the DO-loop termination precedes the DO-statement (statement label %d on line %d)",
					lp->label, lp->lineno);
			else
				errmsg ("a DO-statement must not refer to its own label (label %d)",
					lp->label);
			break;
		}
		/*}}}*/
	}		

	return  lp;
}
/*}}}*/

/*{{{  check_goto_ref (Labelref *refp, Label *lp) -- OK*/
check_goto_ref (refp, lp)
/*---------------------*/
	Labelref *refp;		/* a GOTO reference */
	Label *lp;		/* the target statement */
/*--------------------------------------------------------------*/
/* Called by 'set_part_and_label()' or 'install_label_ref()' to	*/
/* check the validity of a GOTO reference to label 'lp->label'.	*/
/*--------------------------------------------------------------*/
{
	int  from_block = refp->type & LB_BLOCK_MASK;
	int  to_block   = lp->type   & LB_BLOCK_MASK;
	int  to_type    = lp->type   & LB_TYPE_MASK;
	bool  in_goto_stmt = (lineno == refp->lineno);

	switch (to_type) {

	case LB_TARGET :
	case LB_TARGET_OR_DO_TERM :
	case LB_END_IF :
		/*{{{  ok so far -- now check for illegal jump into a block*/
		
		/*--------------------------------------------------------------*/
		/* OK so far -- now check for an illegal jump into a block	*/
		/*--------------------------------------------------------------*/
		if (to_block < from_block  &&  block_in_scope (to_block))
			return;		/* just fine! */
		
		else if (to_block == from_block) {
			/*{{{  give warning if branching to an ENDIF stmt*/
			if (to_type == LB_END_IF) {
				if (in_goto_stmt)
					obsc_warning ("branching to an ENDIF statement (label %d on line %d) from outside the IF construct",
						lp->label, lp->lineno);
				else
					obsc_warning ("branching to an ENDIF statement from outside the IF construct (from line %d)",
						refp->lineno);
			}
			return;		/* OK! */
			
			/*}}}*/
		}
		
		else {
			Block *bp = lookup_block (to_block);
		
			switch (bp->type & BL_TYPE_MASK) {
			case TK_DO :
			case TK_DO_WHILE :
			case TK_DO_FOREVER :
				/*{{{  */
				
				if (in_goto_stmt) {
					if (bp->do_label)
						errmsg ("illegal branch into the range of the \'DO %d\' loop %q (which starts on line %d) at statement label %d on line %d",
							bp->do_label, bp->name, bp->constr_lineno,
							lp->label, lp->lineno);
					else if (bp->name)
						errmsg ("illegal branch into the range of DO-loop %q (which starts on line %d) at statement label %d on line %d",
							bp->name, bp->constr_lineno,
							lp->label, lp->lineno);
					else
						errmsg ("illegal branch into the range of a DO-loop (which starts on line %d) at statement label %d on line %d",
							bp->constr_lineno, lp->label, lp->lineno);
				}
				else {
					if (bp->do_label)
						errmsg ("illegal branch into the range of the \'DO %d\' loop %q from line %d",
							bp->do_label, bp->name, refp->lineno);
					else if (bp->name)
						errmsg ("illegal branch into the range of DO-loop %q from line %d",
							bp->name, refp->lineno);
					else
						errmsg ("illegal branch into the range of a DO-loop from line %d",
							refp->lineno);
				}
				break;
				/*}}}*/
		
			case TK_CASE :
			case TK_DEFAULT :
				/*{{{  */
				if (in_goto_stmt) {
					if (bp->name)
						errmsg ("illegal branch into CASE construct %q (which starts on line %d) at statement label %d on line %d",
							bp->name, bp->constr_lineno, lp->label, lp->lineno);
					else
						errmsg ("illegal branch into a CASE construct (which starts on line %d) at statement label %d on line %d",
							bp->constr_lineno, lp->label, lp->lineno);
				}
				else {
					if (bp->name)
						errmsg ("illegal branch into CASE construct %q from line %d",
							bp->name, refp->lineno);
					else
						errmsg ("illegal branch into a CASE construct from line %d",
							refp->lineno);
				}
				break;
				/*}}}*/
		
			case TK_END_SELECT :
				/*{{{  */
				if (in_goto_stmt)
					errmsg ("illegal branch to an END SELECT statement (label %d on line %d) from outside the CASE construct",
						lp->label, lp->lineno);
				else
					errmsg ("illegal branch to an END SELECT statement from outside the CASE construct (from line %d)",
						refp->lineno);
				break;
				
				/*}}}*/
		
			case TK_IF_CONSTR :
			case TK_ELSE :
			case TK_ELSE_IF :
				/*{{{  */
				if (in_goto_stmt) {
					if (bp->name)
						errmsg ("illegal branch into IF construct %q (which starts on line %d) at statement label %d on line %d",
							bp->name, bp->constr_lineno, lp->label, lp->lineno);
					else
						errmsg ("illegal branch into an IF construct (which starts on line %d) at statement label %d on line %d",
							bp->constr_lineno, lp->label, lp->lineno);
				}
				else {
					if (bp->name)
						errmsg ("illegal branch into IF construct %q from line %d",
							bp->name, refp->lineno);
					else
						errmsg ("illegal branch into an IF construct from line %d",
							refp->lineno);
				}
				break;
				
				/*}}}*/
		
			case TK_WHERE_CONSTR :
			case TK_ELSE_WHERE :
				/*{{{  */
				if (in_goto_stmt)
					errmsg ("illegal branch into a WHERE construct (which starts on line %d) at statement label %d on line %d",
						bp->constr_lineno, lp->label, lp->lineno);
				else
					errmsg ("illegal branch into a WHERE construct from line %d",
						refp->lineno);
				break;
				
				/*}}}*/
			}
			return;
		}
			/*}}}*/

	case LB_FORMAT :
		/*{{{  FORMAT stmt is illegal destination for a branch -- ok*/
		
		/*--------------------------------------------------------------*/
		/* FORMAT statement is an illegal destination for a branch!	*/
		/*--------------------------------------------------------------*/
		
		if (in_goto_stmt)
			errmsg ("illegal branch to a FORMAT statement (label %d on line %d)",
				lp->label, lp->lineno);
		else
			errmsg ("illegal branch to a FORMAT statement from line %d",
				refp->lineno);
		return;
		/*}}}*/

	case LB_ILLEGAL_TARGET :
		/*{{{  other illegal destination for a branch -- ok*/
		
		/*------------------------------------------------------*/
		/* Statement is an illegal destination for a branch.	*/
		/*------------------------------------------------------*/
		
		if (lp->type == LB_ILLEGAL_ELSE_IF) {
			if (in_goto_stmt)
				errmsg ("it is illegal to branch to an ELSE IF statement (label %d on line %d)",
					lp->label, lp->lineno);
			else
				errmsg ("illegal branch to an ELSE IF statement from line %d",
					refp->lineno);
		}
		else if (lp->type == LB_ILLEGAL_ELSE) {
			if (in_goto_stmt)
				errmsg ("it is illegal to branch to an ELSE statement (label %d on line %d)",
					lp->label, lp->lineno);
			else
				errmsg ("illegal branch to an ELSE statement from line %d",
					refp->lineno);
		}
		else if (lp->type == LB_ILLEGAL_CASE) {
			if (in_goto_stmt)
				errmsg ("it is illegal to branch to a CASE statement (label %d on line %d)",
					lp->label, lp->lineno);
			else
				errmsg ("illegal branch to a CASE statement from line %d",
					refp->lineno);
		}
		else {
			if (in_goto_stmt)
				errmsg ("illegal branch to a non-executable statement (label %d on line %d)",
					lp->label, lp->lineno);
			else
				errmsg ("illegal branch to a non-executable statement from line %d",
					refp->lineno);
		}
		return;
		/*}}}*/
	}
}

		/*}}}*/
/*{{{  check_assign_ref (Labelref *refp, Label *lp) -- ok*/
check_assign_ref (refp, lp)
/*-----------------------*/
	Labelref *refp;
	Label *lp;
/*--------------------------------------------------------------*/
/* Called by 'set_part_and_label()' or 'install_label_ref()'	*/
/* to check the validity of assigning label 'lp->label' to a	*/
/* variable in an ASSIGN statement.  This function gives a	*/
/* warning if it is illegal to reference the corresponding	*/
/* statement in *any* context (viz. either in an assigned GOTO	*/
/* or as a format specifier).					*/
/*--------------------------------------------------------------*/
{
	if ( !(lp->type & LB_ILLEGAL_TARGET))
		return;		/* OK! */

	if (lineno == refp->lineno)
		/*------------------------------*/
		/* in the assign statement..	*/
		/*------------------------------*/
		warning ("the statement labelled %d (on line %d) is not a valid reference in either an assigned GOTO or a FORMAT statement",
			lp->label, lp->lineno);
	else
		/*------------------------------*/
		/* in the labelled statement..	*/
		/*------------------------------*/
		warning ("this statement label has been assigned to a variable (on line %d) but is not a valid reference in either an assigned GOTO or a FORMAT statement",
			lp->label, refp->lineno);
}
		/*}}}*/
/*{{{  end_labelled_do_loop (refp, token, n_do_ref) -- OK*/
end_labelled_do_loop (refp, token, n_do_ref)
/*----------------------------------------*/
	Labelref *refp;		/* the DO-statement */
	int  token,  n_do_ref;
/*--------------------------------------------------------------*/
/* Called by 'set_part_and_label()' to process (i.e. check and	*/
/* transform) the terminal statement of a *labelled* DO-loop.	*/
/*--------------------------------------------------------------*/
{	
	bool  inner_do_loop = (n_do_ref == 1);
	int  do_blockno = refp->do_blockno;
	Block  *bp;

	/*{{{  return if this is the DO-stmt itself!!*/
	/*--------------------------------------------------------------*/
	/* This traps the exceptional error where the DO-stmt *itself*	*/
	/* bears the label that it refers to, eg:			*/
	/*								*/
	/*	20   DO 20 I=1,N					*/
	/*	       ....						*/
	/*								*/
	/* In this case an errmsg has already been given by		*/
	/* 'install_label_ref()'.					*/
	/*--------------------------------------------------------------*/
	
	if (refp->lineno == lineno)
		return;
	/*}}}*/
	
	/*{{{  Is the 'DO-block' current?*/
	/*--------------------------------------------------------------*/
	/*	If the 'DO-block' terminated by the current statement	*/
	/* is no longer in scope, this means that it was improperly	*/
	/* nested inside another labelled DO-loop that has already been	*/
	/* processed, and was terminated then (the next-but-one section	*/
	/* performs this action).  If so, give an errmsg and return!	*/
	/*--------------------------------------------------------------*/
	if ( ! block_in_scope (do_blockno)) {
		errmsg ("\'DO %d\' loop beginning on line %d is incorrectly nested",
			stmt_lp->label, refp->lineno);
		return;
	}
	
	
	/*--------------------------------------------------------------*/
	/*	If the 'DO-block' terminated by the DO-term stmt is NOT	*/
	/* the current block, this means that the intervening blocks	*/
	/* are not terminated or are incorrectly nested. Terminate them,*/
	/* giving appropriate errmsgs.					*/
	/*--------------------------------------------------------------*/
	while (current_blockno != do_blockno) {
		/*{{{  terminate current block, giving errmsg*/
		
		bp = lookup_current_block ();
		
		switch (bp->type & BL_TYPE_MASK) {
		case TK_DO :
		case TK_DO_WHILE :
		case TK_DO_FOREVER :
			if (bp->do_label)
				errmsg ("\'DO %d\' loop %q beginning on line %d is not terminated or is incorrectly nested",
					bp->do_label, bp->name, bp->constr_lineno);
			else
				errmsg ("DO-loop %q beginning on line %d is not terminated or is incorrectly nested",
					bp->name, bp->constr_lineno);
			end_current_block ();
			break;
		
		case TK_IF_CONSTR :
		case TK_ELSE_IF :
		case TK_ELSE :
			errmsg ("IF construct %q beginning on line %d is not terminated or is incorrectly nested",
				bp->name, bp->constr_lineno);
			end_current_block ();
			break;
		
		case TK_CASE :
		case TK_DEFAULT :
			errmsg ("CASE construct %q beginning on line %d is not terminated or is incorrectly nested",
				bp->name, bp->constr_lineno);
			end_current_block ();	/* to end the TK_CASE block */
			end_current_block ();	/* to end the TK_END_SELECT block */
			break;
		
		case TK_WHERE_CONSTR :
		case TK_ELSE_WHERE :
			errmsg ("WHERE construct beginning on line %d is not terminated",
				bp->constr_lineno);
			end_current_block ();
			set_unit_part (this_unitp->attrs, EXEC_PART);
			break;
		}
		
		/*--------------------------------------------------------------*/
		/* Adjust 'stmt_lp->type' so that the DO-term is seen to belong	*/
		/* to the innermost DO-loop rather than the un-terminated block	*/
		/* in which it actually appears.				*/
		/*--------------------------------------------------------------*/
		if (inner_do_loop)
			stmt_lp->type = (stmt_lp->type & LB_TYPE_MASK) | current_blockno;
		/*}}}*/
	}
	/*}}}*/

	bp = lookup_block (do_blockno);

	if (bp->name && (token != TK_END_DO)) {
		/*{{{  errmsg -- named DO-loop must be terminated by ENDDO!*/
		errmsg ("named DO-loop %q (beginning on line %d) must be terminated by an ENDDO statement",
			bp->name, bp->constr_lineno);
		if ( !inner_do_loop)
			errmsg ("named DO-loop %q cannot share a DO-termination statement",
				bp->name);
		/*}}}*/
	}
	if (compile)
		write_do_term (bp);

	end_current_block ();
}

	/*}}}*/


/*---------------------------------*/
/* 'newlabel'-generating functions */
/*---------------------------------*/

/*{{{  int  get_newlabel (type) -- ok*/
int  get_newlabel (type)
/*--------------------*/
	int  type;	/* LB_TARGET or LB_DO_TERM */
/*--------------------------------------------------------------*/
/*	This function generates a unique 'newlabel' for a	*/
/* GOTO (and its target statement) or a DO-loop.  'newlabel'	*/
/* means the replacement label that appears in the output	*/
/* program -- remembering that all statement labels in the input*/
/* program are replaced by new ones.  (This is necessary because*/
/* the transformation scheme must introduce extra labels in	*/
/* various contexts, and re-labelling is the only way to	*/
/* guarantee the uniqueness of the extra labels in a one-pass	*/
/* system).							*/
/*	Newlabels for target stmts and DO-loops are generated in*/
/* sequence, starting at 1.  The latest 'newlabel' value is	*/
/* stored in 'std_newlabel'.  DO-loop labels are always		*/
/* multiples of 10 (i.e. 'std_newlabel' is rounded-up to the	*/
/* next multiple of 10), and labels ending in 1 are reserved	*/
/* for use in CONTINUE stmts immediately following a DO-loop	*/
/* termination (to serve as a target for any EXIT stmts that	*/
/* might exist, which are transformed into GOTOs).  The		*/
/* remaining labels (ending 2-9) are used for target stmts.	*/
/*	Format labels are taken from a different sequence,	*/
/* starting at 99999 and decreasing in steps of 1.  They are	*/
/* generated by the function 'change_std_newlabel_to_fmt()'.	*/
/*--------------------------------------------------------------*/
{
	switch (type) {
	case LB_DO_TERM :
		/*----------------------------------------------*/
		/* return 'std_newlabel' rounded-up to the next	*/
		/* multiple of 10.				*/
		/*----------------------------------------------*/
		this_unitp->std_newlabel =
			((this_unitp->std_newlabel + 10) / 10) * 10;

		return  this_unitp->std_newlabel;
		
	default :	/* i.e. LB_TARGET */
		/*----------------------------------------------*/
		/* return 'std_newlabel' incremented by the	*/
		/* minimum value that keeps its final digit in	*/
		/* the range (2-9).				*/
		/*----------------------------------------------*/
		switch (this_unitp->std_newlabel % 10) {
		case 9 :
			return (this_unitp->std_newlabel += 3);
		case 0 :
			return (this_unitp->std_newlabel += 2);
		default :
			return (++this_unitp->std_newlabel);
		}
	}
}
/*}}}*/
/*{{{  int  change_newlabel_to_fmt () -- ok*/
int  change_newlabel_to_fmt ()
/*--------------------------*/
/*--------------------------------------------------------------*/
/*	Function to change a 'std' newlabel to a 'fmt' one,	*/
/* and return the latter!					*/
/*--------------------------------------------------------------*/
{
	/*------------------------------------------------------*/
	/* Restore the 'std_newlabel' to its previous value:	*/
	/*------------------------------------------------------*/
	switch (this_unitp->std_newlabel % 10) {
	case 2 :
		this_unitp->std_newlabel -=
			(this_unitp->std_newlabel == 2) ? 2 : 3;
		break;
	case 1 :	/* shouldn't be possible!! */
		this_unitp->std_newlabel -=
			(this_unitp->std_newlabel == 1) ? 1 : 2;
		break;
	default :
		this_unitp->std_newlabel--;
		break;
	}
	
	/*----------------------------------------------*/
	/* Return 'fmt_newlabel' decremented by 1.	*/
	/*----------------------------------------------*/
	return (--this_unitp->fmt_newlabel);
}	
/*}}}*/
/*{{{  int  remove_newlabel () -- ok*/
int  remove_newlabel ()
/*-------------------*/
/*--------------------------------------------------------------*/
/*	Function to 'remove' a 'std' newlabel that was		*/
/* previously generated.  It restores the 'std' newlabel to its	*/
/* former value, and returns 0 (the 'replacement' label value).	*/
/*--------------------------------------------------------------*/
{
	/*------------------------------------------------------*/
	/* Restore the 'std_newlabel' to its previous value:	*/
	/*------------------------------------------------------*/
	switch (this_unitp->std_newlabel % 10) {
	case 2 :
		this_unitp->std_newlabel -=
			(this_unitp->std_newlabel == 2) ? 2 : 3;
		break;
	case 1 :	/* shouldn't be possible!! */
		this_unitp->std_newlabel -=
			(this_unitp->std_newlabel == 1) ? 1 : 2;
		break;
	default :
		this_unitp->std_newlabel--;
		break;
	}
	
	/*--------------------------------------*/
	/* Return the 'new' newlabel, i.e. 0!	*/
	/*--------------------------------------*/
	return 0;
}	
/*}}}*/

/*{{{  clear_labels () -- ok *** COMMENTED OUT WARNINGS ****/
clear_labels ()
/*-----------*/
/*--------------------------------------------------------------*/
/*	Remove all label records, simultaneously checking that	*/
/* all referenced labels have been defined, and vice versa	*/
/* (unless we're in a type definition or interface block, or	*/
/* we're not in a proc unit, in which case don't bother		*/
/* checking as labels are redundant anyway!)			*/
/*--------------------------------------------------------------*/
{
	Label *lp,  *next_lp;
	Labelref *refp,  *prev_refp;
	bool  check_use = (this_unitp->attrs & PROC_LEVEL)  &&
                          (this_unitp->attrs & UNIT_PROC_TYPE);

	for (lp = this_unitp->first_lp;  lp;  lp = next_lp) {
		/*{{{  if we're in a proc & label is 'real', check its use*/
		/*--------------------------------------------------------------*/
		/* If we're not in a type def or interface block, and it's a	*/
		/* 'real' label (i.e. one appearing in the input program),	*/
		/* check it's use!						*/
		/*--------------------------------------------------------------*/
		if (check_use  &&  lp->label > 0) {
			if (! lp->lineno) {
				/*{{{  errmsg -- label not set*/
				/*--------------------------------------------------------------*/
				/* Label not set, but must have been referenced!  Give errmsg.	*/
				/*--------------------------------------------------------------*/
				int  nrefs;
				
				nrefs = 0;
				for (refp = lp->last_refp;  refp;  refp = refp->prev)
					nrefs++;
				
				if (nrefs == 1)
					begin_errmsg ("label %d not set (referenced on line ", lp->label);
				else
					begin_errmsg ("label %d not set (referenced on lines ", lp->label);
				
				nrefs = 0;
				for (refp = lp->last_refp;  refp;  refp = refp->prev) {
					if (++nrefs == 1)
						cont_errmsg ("%d", refp->lineno);
					else
						cont_errmsg (",%d", refp->lineno);
				}
				
				end_errmsg (")");
				/*}}}*/
			}
		/****************************************************************
		COMMENT THIS OUT FOR NOW, AS IT GIVES SO MANY MISLEADING ERRMSGS
		(AS, E.G., WE DON'T DO ANYTHING WITH LABELS IN I/O STMTS YET!)
		
			else if (! lp->last_refp) 
				warning ("statement label %d (on line %d) is not referenced",
					lp->label, lp->lineno);
		
		*****************************************************************/
		}
		
		/*}}}*/
		/*{{{  clear the label and its associated labelref records*/
		for (refp = lp->last_refp;  refp;  refp = prev_refp) {
			prev_refp = refp->prev;
			free ((char *) refp);
		}
		next_lp = lp->next;
		free ((char *) lp);
		/*}}}*/
	}

	/*------------------------------------------------------*/
	/* 'stmt_lp' & 'stmt_newlabel' must be 0'ed here,	*/
	/* because 'set_part_and_label() may be subsequently	*/
	/* called (e.g. if we're currently closing a unit	*/
	/* abnormally in order to reach a deeper unit that the	*/
	/* current header or 'END' stmt relates to).		*/
	/* If so, 'stmt_lp' must be 0 or 'set_part_and_label()'	*/
	/* will de-reference it--though it's been deleted here!	*/
	/*	However, keep 'stmt_label', because if this is	*/
	/* the END stmt that terminates a CONTAINS section then	*/
	/* the label is re-installed.				*/
	/*------------------------------------------------------*/

	stmt_lp = this_unitp->first_lp = NULL;
	stmt_newlabel = 0;

	/*------------------------------------------------------*/
	/* Re-init 'this_unitp->std_newlabel' & 'fmt_newlabel',	*/
	/* ready for the next proc unit at this level.		*/
	/*------------------------------------------------------*/
	this_unitp->std_newlabel = 0;
	this_unitp->fmt_newlabel = 100000;
}
/*}}}*/
/*}}}*/

/*{{{  <<< block handling functions >>>*/
/*------------------------------*/
/* Block handling functions	*/
/*------------------------------*/
/*{{{  COMMENT -- organisation of the 'block stack' records*/
/******************************************************************

Organisation of the 'block-stack' records:
------------------------------------------
	The 'block stack' is a stack of the block numbers
currently in scope.  It's organised as a list of records, each
containing an *array* of 'BLOCKS_PER_STACK_REC' block numbers
(blstkp->blockno + i) and a pointer to the next record (blstkp->next).
	Incidentally, for efficiency reasons the first 'Blockstack'
record (which will always be at the bottom of the stack) is statically
declared, while any others that are required are 'malloc'ed and 'free'd
when they're done with.
	To demonstrate the order that block numbers are put in the
'block stack' records, say that:

	BLOCKS_PER_STACK_REC = 4;
	n_blocks_in_scope    = 10;

The blocks numbers in this example are randomly distributed but are
in chronological order.  Then, the 'blockstack' records are as follows:

'Blockstack' records:
--------------------
    top_blstkp  ->    1  ->  2  ->  3  ->  '\0'

The first record in the list contains the 'most recent' block numbers.
    
Contents of 'blockstack->(blockno[])':
-------------------------------------
    top_blstkp  ->    1  ->  2  ->  3  ->  '\0'
                      ---------------------
		[0]   18     10     2
		[1]   20     12     6
		[2]   -      15     7
		[3]   -      16     8
		      ---------------------

'current_blockp' points to the number '20' (the 'current_blockno').


Functions in which the 'block-stack' is used / manipulated:
---------------------------------------------------------
The 'block stack' is only used or manipulated in 5 functions:

	begin_block ()
	end_current_block ()
	block_in_scope ()
	block_name_in_scope ()
	innermost_do_loop ()
	
******************************************************************/
/*}}}*/

/*---------------------------------------*/
/* (i) functions to begin and end blocks */
/*---------------------------------------*/

/*{{{  begin_block (token, name, constr_lineno, n_elseif, do_lp, default_bp)--ok*/
begin_block (token, name, constr_lineno, n_elseif, do_lp, default_bp)
/*-----------------------------------------------------------------*/
	int     token;
	char   *name;
	int     constr_lineno;	/* lineno of start of enclosing construct */
	int     n_elseif;	/* only for 'IF' constructs */
	Label  *do_lp;		/* only for 'DO' blocks (labelled or not!) */
	Block  *default_bp;	/* only for 'CASE' blocks! */
/*--------------------------------------------------------------*/
/*	Function to start a new block for a DO-loop or any	*/
/* block construct (IF, SELECT_CASE or WHERE).      'token'	*/
/* indicates the type of block, and is one of the following:	*/
/*								*/
/* for DO-loops:	TK_DO,  TK_DO_WHILE,  TK_DO_FOREVER	*/
/* for IF constr:	TK_IF_CONSTR,  TK_ELSE_IF,  TK_ELSE	*/
/* for CASE constr:	TK_END_SELECT,  TK_CASE,  TK_DEFAULT	*/
/* for WHERE constr:	TK_WHERE_CONSTR,  TK_ELSE_WHERE		*/
/*								*/
/*	The function generates a new 'block record' (*bp) and	*/
/* installs the relevant information, assigns a number to it	*/
/* ('current_blockno'), & puts this number on the 'block stack'.*/
/*	Note that this function ***handles all new blocks.***	*/
/*--------------------------------------------------------------*/
{
	int  new_elem;
	Blockrec  *blrecp;
	Block  *bp;

	current_blockno = ++n_blocks;
	
	/*{{{  set 'bp' to point to the next new 'Block' record*/
	/*------------------------------------------------------*/
	/* Set 'bp' to point to the next new 'block record'.	*/
	/*------------------------------------------------------*/
	
	new_elem = (n_blocks-1) % BLOCKS_PER_REC;
	
	if ( !first_blrecp)		/* no blocks at all yet! */
		blrecp = first_blrecp = (Blockrec *) ecalloc (sizeof (Blockrec));
	else {
		for (blrecp = first_blrecp;  blrecp->next;  blrecp = blrecp->next)
			;
		if (new_elem == 0) {
			/*----------------------------------------------*/
			/* add new block record to the end of the list	*/
			/*----------------------------------------------*/
			blrecp->next = (Blockrec *) ecalloc (sizeof (Blockrec));
			blrecp = blrecp->next;
		}
	}
	
	bp = (blrecp->block) + new_elem;	/* pointer to a block */
	/*}}}*/

	/*{{{  enter block information in '*bp'*/
	/*----------------------------------------------*/
	/* Enter general block information in '*bp':	*/
	/*----------------------------------------------*/
	bp->type = (token == TK_DO_WHILE) ? (token | BL_EXIT_REF) : token;
			/* the transformation of 'DO_WHILE' loops	*/
			/* always involves an EXIT branch if the	*/
			/* WHILE condition is not satisfied!		*/
	bp->name = name;
	bp->block_lineno = lineno;
	bp->constr_lineno = constr_lineno ? constr_lineno : lineno;
	
	/*------------------------------------------------------*/
	/* The following information is specific to particular	*/
	/* constructs:						*/
	/*------------------------------------------------------*/
	switch (token) {
	case  TK_CASE :
		bp->default_bp = default_bp;
		break;
	
	case  TK_DEFAULT :
		bp->default_bp = bp;	/* this is the 'CASE DEFAULT' block! */
		break;
	
	case  TK_DO :
	case  TK_DO_WHILE :
	case  TK_DO_FOREVER :
		/*------------------------------------------------------*/
		/* N.B. 'do_lp' is always defined for DO-loops, even if	*/
		/* they're unlabelled.  In that case, 'do_lp' points to	*/
		/* a record containing the value LB_UNLABELLED_DO	*/
		/* (probably -1!) for 'do_lp->label'.  The only use of	*/
		/* this record is that 'do_lp->last_refp' points to the	*/
		/* record that we **actually** need (which holds	*/
		/* 'newlabel', 'lineno', etc.)				*/
		/*------------------------------------------------------*/
		bp->do_label = (do_lp->label == LB_UNLABELLED_DO) ? 0 : do_lp->label;
	
		bp->do_refp = do_lp->last_refp;
		bp->do_refp->do_blockno = current_blockno;
	
		break;
	
	case  TK_IF :
	case  TK_ELSE_IF :
	case  TK_ELSE :
		bp->do_label = n_elseif;
		break;
	}
	/*}}}*/
	
	/*{{{  install 'current_blockno' in 'Blockstack'*/
	
	/*--------------------------------------------------------------*/
	/* Install block number in 'Blockstack'.			*/
	/* (This involves modifying 'n_blocks_in_scope','current_blockp'*/
	/* and possibly 'top_blstkp'.)					*/
	/*--------------------------------------------------------------*/
	
	if (n_blocks_in_scope == 0)
		/*------------------------------------------------------*/
		/* Put the block number in the (static) first		*/
		/* 'Blockstack' record, pointed to by 'top_blstkp'.	*/
		/*------------------------------------------------------*/
		current_blockp = top_blstkp->blockno;
	
	else if ((n_blocks_in_scope % BLOCKS_PER_STACK_REC) == 0) {
		/*------------------------------------------------------*/
		/* Create a new 'Blockstack' record and add it to the	*/
		/* *head* of the list  (cf, the 'Blockrec' record is	*/
		/* linked to the tail of its list!)			*/
		/*------------------------------------------------------*/
		Blockstack  *blstkp;
	
		blstkp = (Blockstack *) emalloc (sizeof (Blockstack));
		blstkp->next = top_blstkp;
		top_blstkp = blstkp;
		current_blockp  = blstkp->blockno;
	}
	else
		++current_blockp;		
	
		
	*current_blockp = current_blockno;
	++n_blocks_in_scope;
	/*}}}*/
}

/*}}}*/
/*{{{  end_current_block () -- ok*/
end_current_block ()
/*----------------*/
/*--------------------------------------------------------------*/
/*	Terminates the block currently in scope (i.e. removes	*/
/* it from the 'blockstack' but *does* not delete the 'block	*/
/* record').  Removal of a block number involves decrementing	*/
/* 'n_blocks_in_scope', changing the pointer 'current_blockp'	*/
/* and the value 'current_blockno' that it points to, and	*/
/* possibly deallocating a 'block stack' record -- it is *not*	*/
/* necessary to replace the old value of '*current_blockp' by 0.*/
/*	This is called by functions such as 'begin_else_block()'*/
/* etc, which have already checked that a current block exists,	*/
/* so no checking is necessary.					*/
/*--------------------------------------------------------------*/
{
	--n_blocks_in_scope;

	if (n_blocks_in_scope == 0)
		/*----------------------------------------------*/
		/* remove last block number from the bottom	*/
		/* 'Blockstack' record, but *don't* free the	*/
		/* record (since it's static!) and don't change	*/
		/* 'top_blstkp'.				*/
		/*----------------------------------------------*/
		current_blockp = NULL;

	else if ((n_blocks_in_scope % BLOCKS_PER_STACK_REC) == 0) {
		/*----------------------------------------------*/
		/* remove the last block number in the current	*/
		/* 'Blockstack' record.  This is *not* the	*/
		/* record, so remove it.			*/
		/*----------------------------------------------*/
		Blockstack  *blstkp = top_blstkp;

		top_blstkp = blstkp->next;
		free ((char *) blstkp);
		current_blockp = (top_blstkp->blockno) + BLOCKS_PER_STACK_REC - 1;
	}
	else
		--current_blockp;

		
	current_blockno = current_blockp ? *current_blockp : 0;
}
/*}}}*/

/*{{{  begin_else_if_block (char *name)  -- ok*/
begin_else_if_block (name)
/*----------------------*/
	char *name;	/* the name referenced in the ELSEIF stmt (if any) */
/*--------------------------------------------------------------*/
/*	Called by 'yyparse' on reading an ELSEIF statement,	*/
/* to begin an 'ELSEIF block'.  It only does so if the		*/
/* statement is valid (i.e. if the current block is an IF or	*/
/* ELSEIF block, but NOT an ELSE block) -- if it's invalid, an	*/
/* error message is given and the function returns immediately.	*/
/* If the statement bears a block name (i.e 'name' <> NULL),	*/
/* it is checked that this is the actual name of the IF		*/
/* construct --	 if not, an error message is given but the	*/
/* block is still created.					*/
/*	If the ELSEIF statement is legal, the previous block is	*/
/* ended and the new 'ELSEIF block' started. The (correct) name	*/
/* of the construct, if any, is installed in the new block.	*/
/*--------------------------------------------------------------*/
{
	int    constr_lineno,  n_elseif;
	Block *bp = lookup_current_block ();
	
	/*{{{  return immediately if we're not in an IF constr.*/
	/*--------------------------------------------------------------*/
	/*	Return immediately if we're not in an IF construct,	*/
	/* or if the current block is an ELSE block (as it should be	*/
	/* the last block of the IF construct!)				*/
	/*	Otherwise, find the number of previous ELSEIF blocks	*/
	/* so we can evaluate 'n_elseif' (to store in 'bp->do_label'!).	*/
	/*--------------------------------------------------------------*/
	if (this_unitp->attrs & WHERE_PART) {
		errmsg ("only assignments are allowed in a WHERE construct (which begins on line %d) - missing \'ENDWHERE\' is assumed",
			bp->constr_lineno);
		end_current_block ();
		set_unit_part (this_unitp->attrs, EXEC_PART);
		bp = lookup_current_block ();
	}
	
	if (!bp) {
		errmsg ("misplaced ELSEIF statement - not in an IF construct");
		return;
	}
	
	switch (bp->type & BL_TYPE_MASK) {
	case TK_DO :
	case TK_DO_WHILE :
	case TK_DO_FOREVER :
		errmsg ("misplaced ELSEIF statement - located in a DO-loop (beginning on line %d), not an IF construct",
			bp->constr_lineno);
		return;
	
	case TK_CASE :
	case TK_DEFAULT :
		errmsg ("misplaced ELSEIF statement - located in a CASE construct (beginning on line %d), not an IF construct",
			bp->constr_lineno);
		return;
	
	case TK_ELSE :
		errmsg ("the ELSE block (beginning on line %d) must be the final block of the IF construct",
			bp->block_lineno);
		return;
	
	default :	/* 'bp' is an 'IF' or 'ELSEIF' block */
	
		constr_lineno = bp->constr_lineno;
		n_elseif      = bp->do_label + 1;
		break;
	}
	/*}}}*/
	
	/*{{{  check that the block name (if any) is correct*/
	/*--------------------------------------------------------------*/
	/* Check that the block name (if any) is correct.		*/
	/* N.B. The following tests assume that a given name is stored	*/
	/* uniquely, i.e. it cannot be duplicated in different storage	*/
	/* locations!							*/
	/*--------------------------------------------------------------*/
	if (name  &&  name != bp->name) {
		if (!bp->name)
			errmsg ("ELSEIF statement must not specify a name as it belongs to an unnamed IF construct (beginning on line %d)",
			           constr_lineno);
		else 
			errmsg ("incorrect construct name %q - the enclosing construct (beginning on line %d) is called %q",
			           name, constr_lineno, bp->name);
		name = bp->name;
	}
	/*}}}*/
	
	end_current_block ();
	begin_block (TK_ELSE_IF, name, constr_lineno, n_elseif,
	                         NULL_Label,  NULL_Block);
}
/*}}}*/
/*{{{  begin_else_block (char *name) -- ok*/
begin_else_block (name)
/*-------------------*/
	char *name;	/* the name referenced in the ELSE stmt (if any) */
/*--------------------------------------------------------------*/
/*	Called by 'yyparse' on reading an ELSE statement,	*/
/* to begin an 'ELSE block'.  It only does so if the ELSE	*/
/* statement is valid (i.e. if the current block is an IF or	*/
/* ELSEIF block, but NOT an ELSE block) -- if it's invalid, an	*/
/* error message is given and the function returns immediately.	*/
/* If the statement bears a block name (i.e 'name' <> NULL),	*/
/* it is checked that this is the actual name of the IF		*/
/* construct --	 if not, an error message is given but the	*/
/* block is still created.					*/
/*	If the ELSE statement is legal, the previous block is	*/
/* ended and the new 'ELSE block' started. The (correct) name	*/
/* of the construct, if any, is installed in the new block.	*/
/*--------------------------------------------------------------*/
{
	int    constr_lineno,  n_elseif;
	Block *bp = lookup_current_block ();

	/*{{{  return immediately if we're not in an IF constr.*/
	/*--------------------------------------------------------------*/
	/*	Return immediately if we're not in an IF construct.	*/
	/* or if the current block is an ELSE block (as there can only	*/
	/* be one ELSE block!)						*/
	/*	Otherwise, find the number of previous ELSEIF blocks	*/
	/* so we can evaluate 'n_elseif' (to store in 'bp->do_label'!).	*/
	/*--------------------------------------------------------------*/
	if (this_unitp->attrs & WHERE_PART) {
		errmsg ("only assignments are allowed in a WHERE construct (which begins on line %d) - missing \'ENDWHERE\' is assumed",
			bp->constr_lineno);
		end_current_block ();
		set_unit_part (this_unitp->attrs, EXEC_PART);
		bp = lookup_current_block ();
	}
	
	if (!bp) {
		errmsg ("misplaced ELSE statement - not in an IF construct");
		return;
	}
	
	switch (bp->type & BL_TYPE_MASK) {
	case TK_DO :
	case TK_DO_WHILE :
	case TK_DO_FOREVER :
		errmsg ("misplaced ELSE statement - located in a DO-loop (beginning on line %d), not an IF construct",
			bp->constr_lineno);
		return;
	
	case TK_CASE :
	case TK_DEFAULT :
		errmsg ("misplaced ELSE statement - located in a CASE construct (beginning on line %d), not an IF construct",
			bp->constr_lineno);
		return;
	
	case TK_ELSE :
		errmsg ("ELSE block already given (beginning on line %d)",
			bp->block_lineno);
		return;
	
	default :	/* 'bp' is an 'IF' or 'ELSEIF' block */
	
		constr_lineno = bp->constr_lineno;
		n_elseif      = bp->do_label;
		break;
	}
	/*}}}*/

	/*{{{  check that the block name (if any) is correct*/
	/*--------------------------------------------------------------*/
	/* Check that the block name (if any) is correct.		*/
	/* N.B. The following tests assume that a given name is stored	*/
	/* uniquely, i.e. it cannot be duplicated in different storage	*/
	/* locations!							*/
	/*--------------------------------------------------------------*/
	if (name  &&  name != bp->name) {
		if (!bp->name)
			errmsg ("ELSE statement must not specify a name as it belongs to an unnamed IF construct (beginning on line %d)",
			            constr_lineno);
		else 
			errmsg ("incorrect construct name %q - the enclosing construct (beginning on line %d) is called %q",
			           name, constr_lineno, bp->name);
		name = bp->name;
	}
	/*}}}*/
	
	end_current_block ();
	begin_block (TK_ELSE, name, constr_lineno, n_elseif,
	                      NULL_Label,  NULL_Block);
}
/*}}}*/

/*{{{  begin_case_block (char *name)  -- ok*/
begin_case_block (name)
/*-------------------*/
	char *name;	/* the name referenced in the CASE stmt (if any) */
/*--------------------------------------------------------------*/
/*	Called by 'yyparse' on reading an CASE statement,	*/
/* to begin a 'CASE block'.  It only does so if the CASE	*/
/* statement is valid (i.e. if the current block is a CASE or	*/
/* DEFAULT block, signalling that we're in a SELECT CASE	*/
/* construct) -- if not, an error message is given and the	*/
/* function returns immediately.  If the CASE statement bears a	*/
/* block name (i.e 'name' <> NULL), it is checked that this is	*/
/* the actual name of the CASE construct -- if not, an error	*/
/* message is given but the block is still created.		*/
/*	If the CASE statement is legal, the previous block is	*/
/* ended and the new 'CASE block' started. The (correct) name	*/
/* of the construct, if any, is installed in the new block.	*/
/*--------------------------------------------------------------*/
{
	Block *bp = lookup_current_block ();
	
	/*{{{  return immediately if we're not in a CASE constr.*/
	/*------------------------------------------------------*/
	/* Return immediately if we're not in a CASE construct.	*/
	/*------------------------------------------------------*/
	if (this_unitp->attrs & WHERE_PART) {
		errmsg ("only assignments are allowed in a WHERE construct (which begins on line %d) - missing \'ENDWHERE\' is assumed",
			bp->constr_lineno);
		end_current_block ();
		set_unit_part (this_unitp->attrs, EXEC_PART);
		bp = lookup_current_block ();
	}
	
	if (!bp) {
		errmsg ("misplaced CASE statement - not in a CASE construct");
		return;
	}
	
	switch (bp->type & BL_TYPE_MASK) {
	case TK_DO :
	case TK_DO_WHILE :
	case TK_DO_FOREVER :
		errmsg ("misplaced CASE statement - located in a DO-loop (beginning on line %d), not a CASE construct",
			bp->constr_lineno);
		return;
	
	case TK_IF_CONSTR :
	case TK_ELSE_IF :
	case TK_ELSE :
		errmsg ("misplaced CASE statement - located in an IF construct (beginning on line %d), not a CASE construct",
			bp->constr_lineno);
		return;
	}
	/*}}}*/
	
	/*{{{  check that the block name (if any) is correct*/
	/*--------------------------------------------------------------*/
	/* Check that the block name (if any) is correct.		*/
	/* N.B. The following tests assume that a given name is stored	*/
	/* uniquely, i.e. it cannot be duplicated in different storage	*/
	/* locations!							*/
	/*--------------------------------------------------------------*/
	if (name  &&  name != bp->name) {
		if (!bp->name)
			errmsg ("CASE statement must not specify a name as it belongs to an unnamed CASE construct (beginning on line %d)",
				bp->constr_lineno);
		else
			errmsg ("incorrect construct name %q - the enclosing construct (beginning on line %d) is called %q",
				name, bp->constr_lineno, bp->name);
	}
	/*}}}*/
	
	end_current_block ();
	begin_block (TK_CASE,  bp->name,  bp->constr_lineno,  0,
	                       NULL_Label,  bp->default_bp);
}
/*}}}*/
/*{{{  begin_default_block (char *name) -- ok*/
begin_default_block (name)
/*----------------------*/
	char *name;	/* the name referenced in the CASE stmt (if any) */
/*--------------------------------------------------------------*/
/*	Called by 'yyparse' on reading an CASE DEFAULT		*/
/* statement, to begin a 'DEFAULT block'.  It only does so if	*/
/* the CASE statement is valid (i.e. if the current block is a	*/
/* CASE block, signalling that we're in a CASE construct, and	*/
/* a DEFAULT block has not already appeared) -- if not, an	*/
/* error message is given and the function returns immediately.	*/
/* If the CASE statement bears a block name (name <> NULL), it	*/
/* is checked that this is the actual name of the CASE construct*/
/* -- if not, an error message is given but the block is still	*/
/* created.							*/
/*	If the CASE statement is legal, the previous block is	*/
/* ended and the new 'DEFAULT block' started. The (correct) name*/
/* of the construct, if any, is installed in the new block.	*/
/*--------------------------------------------------------------*/
{
	Block *bp = lookup_current_block ();
	
	/*{{{  return immediately if we're not in a CASE constr.*/
	/*------------------------------------------------------*/
	/* Return immediately if we're not in a CASE construct.	*/
	/*------------------------------------------------------*/
	if (this_unitp->attrs & WHERE_PART) {
		errmsg ("only assignments are allowed in a WHERE construct (which begins on line %d) - missing \'ENDWHERE\' is assumed",
			bp->constr_lineno);
		end_current_block ();
		set_unit_part (this_unitp->attrs, EXEC_PART);
		bp = lookup_current_block ();
	}
	
	if (!bp) {
		errmsg ("misplaced CASE statement - not in a CASE construct");
		return;
	}
	
	switch (bp->type & BL_TYPE_MASK) {
	case TK_DO :
	case TK_DO_WHILE :
	case TK_DO_FOREVER :
		errmsg ("misplaced CASE statement - located in a DO-loop (beginning on line %d), not a CASE construct",
			bp->constr_lineno);
		return;
	
	case TK_IF_CONSTR :
	case TK_ELSE_IF :
	case TK_ELSE :
		errmsg ("misplaced CASE statement - located in an IF construct (beginning on line %d), not a CASE construct",
			bp->constr_lineno);
		return;
	}
	/*}}}*/
	
	/*{{{  check that the block name (if any) is correct*/
	/*--------------------------------------------------------------*/
	/* Check that the block name (if any) is correct.		*/
	/* N.B. The following tests assume that a given name is stored	*/
	/* uniquely, i.e. it cannot be duplicated in different storage	*/
	/* locations!							*/
	/*--------------------------------------------------------------*/
	if (name  &&  name != bp->name) {
		if (!bp->name)
			errmsg ("CASE statement must not specify a name as it belongs to an unnamed CASE construct (beginning on line %d)",
				bp->constr_lineno);
		else
			errmsg ("incorrect construct name %q - the enclosing construct (beginning on line %d) is called %q",
				name, bp->constr_lineno, bp->name);
	}
	/*}}}*/
	
	/*{{{  return if there has already been a CASE DEFAULT block*/
	/*--------------------------------------------------------------*/
	/* Return if there has already been a CASE DEFAULT block (as	*/
	/* there can only be one such block per CASE construct!)	*/
	/*--------------------------------------------------------------*/
	if (bp->default_bp) {
		errmsg ("CASE DEFAULT block already given (beginning on line %d)",
			bp->default_bp->block_lineno);
		return;
	}
	/*}}}*/

	end_current_block ();
	begin_block (TK_DEFAULT,  bp->name,  bp->constr_lineno,  0,
	                          NULL_Label,  NULL_Block);
}
/*}}}*/
/*{{{  begin_else_where_block () -- ok*/
begin_else_where_block ()
/*---------------------*/
/*--------------------------------------------------------------*/
/*	Called by 'yyparse' on reading an ELSEWHERE statement,	*/
/* to begin an 'ELSEWHERE block'.  It only does so if the	*/
/* statement is valid, i.e. if the current block is a WHERE	*/
/* block -- if not, an error message is given and the function	*/
/* returns immediately.						*/
/*	If the ELSEWHERE statement is legal, the previous block	*/
/* is ended and the new 'ELSEWHERE block' started.		*/
/*--------------------------------------------------------------*/
{
	Block *bp = lookup_current_block ();

	/*{{{  return immediately if we're not in a WHERE block.*/
	/*------------------------------------------------------*/
	/* Return immediately if we're not in a WHERE block.	*/
	/*------------------------------------------------------*/
	if (!bp) {
		errmsg ("misplaced ELSEWHERE statement - not in a WHERE construct");
		return;
	}
	
	switch (bp->type & BL_TYPE_MASK) {
	case TK_WHERE_CONSTR :		/* OK! */
		break;
		
	case TK_ELSE_WHERE :
		errmsg ("ELSEWHERE block already given (beginning on line %d)", bp->block_lineno);
		return;
	
	case TK_DO :
	case TK_DO_WHILE :
	case TK_DO_FOREVER :
		errmsg ("misplaced ELSEWHERE statement - located in a DO-loop (beginning on line %d), not a WHERE construct",
			bp->constr_lineno);
		return;
	
	case TK_IF_CONSTR :
	case TK_ELSE_IF :
	case TK_ELSE :
		errmsg ("misplaced ELSEWHERE statement - located in an IF construct (beginning on line %d), not a WHERE construct",
			bp->constr_lineno);
		return;
	
	case TK_CASE :
	case TK_DEFAULT :
		errmsg ("misplaced ELSEWHERE statement - located in a CASE construct (beginning on line %d), not a WHERE construct",
			bp->constr_lineno);
		return;
	
	}
	/*}}}*/

	end_current_block ();
	begin_block (TK_ELSE_WHERE,  NULL_str,  bp->constr_lineno,  0,
	                             NULL_Label,  NULL_Block);
}
/*}}}*/

/*{{{  int  end_if_block (name) -- ok*/
int  end_if_block (name)
/*--------------------*/
	char *name;	/* the name referenced in the ENDIF stmt (if any) */
/*--------------------------------------------------------------*/
/*	Called by 'yyparse' on reading an ENDIF statement.  The	*/
/* statement is only valid if the current block is part of an	*/
/* IF construct (i.e. it is an IF, ELSE_IF or ELSE block) --	*/
/* if not, an error message is given and the function returns	*/
/* immediately.  If this test is passed, the function also	*/
/* checks that 'name', the name following the ENDIF keyword,	*/
/* is the name of the current IF construct (or NULL if the	*/
/* construct is unnamed) -- if not, an error message is given	*/
/* but the IF construct is still terminated.			*/
/*	'Termination' of the IF construct consists of terminat-	*/
/* ing the block currently in scope (i.e. removing it from the	*/
/* 'blockstack' but NOT deleting its 'block record').		*/
/*	The number of 'ELSEIF' blocks is returned.		*/
/*--------------------------------------------------------------*/
{
	int    n_elseif;
	Block *bp = lookup_current_block ();
	
	/*{{{  return immediately if we're not in an IF constr.*/
	/*------------------------------------------------------*/
	/* Return immediately if we're not in an IF construct.	*/
	/*------------------------------------------------------*/
	if (this_unitp->attrs & WHERE_PART) {
		errmsg ("only assignments are allowed in a WHERE construct (which begins on line %d) - missing \'ENDWHERE\' is assumed",
			bp->constr_lineno);
		end_current_block ();
		set_unit_part (this_unitp->attrs, EXEC_PART);
		bp = lookup_current_block ();
	}
	
	if (!bp) {
		errmsg ("misplaced ENDIF statement - not in an IF construct");
		return  0;
	}
	
	switch (bp->type & BL_TYPE_MASK) {
	case TK_DO :
	case TK_DO_WHILE :
	case TK_DO_FOREVER :
		errmsg ("misplaced ENDIF statement - located in a DO-loop (beginning on line %d), not an IF construct",
			bp->constr_lineno);
		return  0;
	
	case TK_CASE :
	case TK_DEFAULT :
		errmsg ("misplaced ENDIF statement - located in a CASE construct (beginning on line %d), not an IF construct",
			bp->constr_lineno);
		return  0;
	
	default :	/* 'bp' is an 'IF', 'ELSEIF' or 'ELSE' block */
	
		n_elseif = bp->do_label;	/* ...the # of 'ELSEIF' blocks */
		break;
	}
	/*}}}*/

	/*------------------------------------------------------*/
	/* Check that 'name' exactly matches that of the IF	*/
	/* construct (which may be NULL!).			*/
	/*  N.B. The following tests assume that a given name	*/
	/* is stored uniquely, i.e. it is not duplicated in	*/
	/* different storage locations!				*/
	/*------------------------------------------------------*/
	if (name != bp->name) {
		if (bp->name)
			errmsg ("ENDIF statement must specify the name %q of the IF construct that it terminates (which begins on line %d)",
				bp->name, bp->constr_lineno);
		else 
			errmsg ("ENDIF statement must not specify a name as it terminates an unnamed IF construct (which begins on line %d)",
				bp->constr_lineno);
	}
	
	end_current_block ();

	return  n_elseif;	/* ...the # of 'ELSEIF' blocks */
}
/*}}}*/
/*{{{  bool  end_case_block (char *name) -- ok*/
bool  end_case_block (name)
/*-----------------------*/
	char *name;	/* the name referenced in the END SELECT stmt (if any) */
/*--------------------------------------------------------------*/
/*	Called by 'yyparse' on reading an END SELECT statement.	*/
/* The statement is only valid if the current block is part of	*/
/* CASE construct (i.e. it is a CASE or DEFAULT block) --*/
/* if not, an error message is given and the function returns	*/
/* immediately.  If this test is passed, the function also	*/
/* checks that 'name', the name following the ENDSELECT keyword,*/
/* is the name of the current construct (or NULL if the		*/
/* construct is unnamed) -- if not, an error message is given	*/
/* but the CASE construct is still terminated.		*/
/*	'Termination' of the CASE construct consists of	*/
/* terminating the block currently in scope (i.e. removing it	*/
/* from the 'blockstack' but NOT deleting its 'block record').	*/
/*	The function returns a boolean result indicating whether*/
/* the block termination is performed (i.e. whether the END	*/
/* SELECT statement was legal).					*/
/*	Note that there is also an 'underlying' END_SELECT block*/
/* that must also be 'terminated'.  If this function returns a	*/
/* TRUE result, 'yyparse' subsequently calls 'end_current_block'*/
/* to terminate the latter block, after processing labels, etc.	*/
/*--------------------------------------------------------------*/
{
	Block *bp = lookup_current_block ();
	
	/*{{{  return immediately if we're not in an CASE constr.*/
	/*------------------------------------------------------*/
	/* Return immediately if we're not in a CASE construct.	*/
	/*------------------------------------------------------*/
	if (this_unitp->attrs & WHERE_PART) {
		errmsg ("only assignments are allowed in a WHERE construct (which begins on line %d) - missing \'ENDWHERE\' is assumed",
			bp->constr_lineno);
		end_current_block ();
		set_unit_part (this_unitp->attrs, EXEC_PART);
		bp = lookup_current_block ();
	}
	
	if (!bp) {
		errmsg ("misplaced END SELECT statement - not in a CASE construct");
		return FALSE;
	}
	
	switch (bp->type & BL_TYPE_MASK) {
	case TK_DO :
	case TK_DO_WHILE :
	case TK_DO_FOREVER :
		errmsg ("misplaced END SELECT statement - located in a DO-loop (beginning on line %d), not a CASE construct",
			bp->constr_lineno);
		return FALSE;
	
	case TK_IF_CONSTR :
	case TK_ELSE_IF :
	case TK_ELSE :
		errmsg ("misplaced END SELECT statement - located in an IF construct (beginning on line %d), not a CASE construct",
			bp->constr_lineno);
		return FALSE;
	}
	/*}}}*/

	/*------------------------------------------------------*/
	/* Check that 'name' exactly matches that of the	*/
	/* CASE construct (which may be NULL!).			*/
	/*  N.B. The following tests assume that a given name	*/
	/* is stored uniquely, i.e. it is not duplicated in	*/
	/* different storage locations!				*/
	/*------------------------------------------------------*/
	if (name != bp->name) {
		if (bp->name)
			errmsg ("END SELECT statement must specify the name %q of the CASE construct that it terminates (which begins on line %d)",
				bp->name, bp->constr_lineno);
		else 
			errmsg ("END SELECT statement must not specify a name as it terminates an unnamed CASE construct (which begins on line %d)",
				bp->constr_lineno);
	}
	
	end_current_block ();
	return  TRUE;
}
/*}}}*/
/*{{{  end_unlabelled_do_loop (char *name) -- ok*/
end_unlabelled_do_loop (name)
/*-------------------------*/
	char *name;	/* the name referenced in the ENDDO stmt (if any) */
/*--------------------------------------------------------------*/
/*	Called by 'yyparse' on reading an ENDDO statement that	*/
/* is either unlabelled or bears a label that isn't referenced	*/
/* by any DO-statement.  (An ENDDO belonging to a labelled DO-	*/
/* loop is processed by 'end_labelled_do_loop').		*/
/*	The statement is only valid if the current block *is* a	*/
/* DO-loop (i.e. it is a DO, DO_WHILE or DO_FOREVER block) --	*/
/* if not, an error message is given and the function returns	*/
/* immediately.  If this test is passed, the function also	*/
/* checks that 'name', the name following the ENDDO keyword,	*/
/* is the name of the current DO-loop (or NULL if the DO-loop	*/
/* is unnamed) -- if not, an error message is given but the	*/
/* DO-loop is still terminated.					*/
/*	'Termination' of the DO-loop consists of terminating	*/
/* the block currently in scope (i.e. removing it from the	*/
/* 'blockstack' but NOT deleting its 'block record').		*/
/*--------------------------------------------------------------*/
{
	Block *bp = lookup_current_block ();

	/*{{{  return immediately if we're not in a DO-loop.*/
	/*--------------------------------------------------------------*/
	/* Return immediately if we're not in a DO-loop.		*/
	/* N.B. We can't be in a WHERE construct, as			*/
	/* 'set_part_and_label ()' has already been called and would	*/
	/* have terminated the WHERE block with an errmsg!		*/
	/*--------------------------------------------------------------*/
	if (!bp) {
		errmsg ("misplaced ENDDO statement - not in a DO-loop");
		return;
	}
	
	switch (bp->type & BL_TYPE_MASK) {
	case TK_IF_CONSTR :
	case TK_ELSE_IF :
	case TK_ELSE :
		errmsg ("misplaced ENDDO statement - located in an IF construct (beginning on line %d), not a DO-loop",
			bp->constr_lineno);
		return;
	
	case TK_CASE :
	case TK_DEFAULT :
		errmsg ("misplaced ENDDO statement - located in a CASE construct (beginning on line %d), not a DO-loop",
			bp->constr_lineno);
		return;
	}
	/*}}}*/

	/*{{{  check whether the DO-stmt was labelled -- if so, give errmsg*/
	
	/*--------------------------------------------------------------*/
	/*	Check whether the DO-statement referenced a label.	*/
	/* If so, then the ENDDO statement doesn't bear the required	*/
	/* label (either it bears no label or the wrong one) -- since	*/
	/* otherwise 'set_part_and_label' and 'end_labelled_do_loop()'	*/
	/* would process it and this function wouldn't be called.	*/
	/* Give an errmsg.						*/
	/*--------------------------------------------------------------*/
	if (bp->do_label)
		errmsg ("ENDDO statement must bear the label %d used in the corresponding \'DO %d\' statement (on line %d)",
			bp->do_label, bp->do_label, bp->constr_lineno);
	/*}}}*/
	
	/*{{{  check 'name'*/
	/*--------------------------------------------------------------*/
	/* Check that 'name' exactly matches that of the DO-loop (which	*/
	/* may be NULL!).						*/
	/* N.B. The following tests assume that a given name is stored	*/
	/* uniquely, i.e. it is not duplicated in different storage	*/
	/* locations!							*/
	/*--------------------------------------------------------------*/
	if (name != bp->name) {
		if (bp->name)
			errmsg ("ENDDO statement must specify the name %q of the DO-loop that it terminates (which begins on line %d)",
				bp->name, bp->constr_lineno);
		else 
			errmsg ("ENDDO statement must not specify a name as it terminates an unnamed DO-loop (which begins on line %d)",
				bp->constr_lineno);
	}
	/*}}}*/

	if (compile)
		write_do_term (bp);

	end_current_block ();
}
/*}}}*/
/*{{{  end_where_block () -- ok*/
end_where_block ()
/*--------------*/
/*--------------------------------------------------------------*/
/*	Called to end a WHERE or ELSEWHERE block, either by	*/
/* 'yyparse' on reading an ENDWHERE statement, or by		*/
/* 'set_part_and_label ()' or one of the block-handling routines*/
/* when (unit_part == WHERE_PART) and a non-assignment is read.	*/
/* In the first case, the ENDWHERE statement is only valid if	*/
/* we're in a WHERE construct (i.e. it the current block is a	*/
/* WHERE or ELSEWHERE block) -- if not, an error message is	*/
/* given and the function returns immediately.			*/
/*	'Termination' of the WHERE construct consists of	*/
/* terminating the block currently in scope (i.e. removing it	*/
/* from the 'blockstack' but NOT deleting its 'block record'),	*/
/* and changing 'unit_part' from WHERE_PART to EXEC_PART.	*/
/*--------------------------------------------------------------*/
{
	Block *bp = lookup_current_block ();

	/*{{{  return immediately if we're not in a WHERE block.*/
	/*------------------------------------------------------*/
	/* Return immediately if we're not in a WHERE block.	*/
	/*------------------------------------------------------*/
	if (!bp) {
		errmsg ("misplaced ENDWHERE statement - not in a WHERE construct");
		return;
	}
	
	switch (bp->type & BL_TYPE_MASK) {
	case TK_WHERE_CONSTR :
	case TK_ELSE_WHERE :
		break;		/* OK! */
		
	case TK_DO :
	case TK_DO_WHILE :
	case TK_DO_FOREVER :
		errmsg ("misplaced ENDWHERE statement - located in a DO-loop (beginning on line %d), not a WHERE construct",
			bp->constr_lineno);
		return;
	
	case TK_IF_CONSTR :
	case TK_ELSE_IF :
	case TK_ELSE :
		errmsg ("misplaced ENDWHERE statement - located in an IF construct (beginning on line %d), not a WHERE construct",
			bp->constr_lineno);
		return;
	
	case TK_CASE :
	case TK_DEFAULT :
		errmsg ("misplaced ENDWHERE statement - located in a CASE construct (beginning on line %d), not a WHERE construct",
			bp->constr_lineno);
		return;
	}
	/*}}}*/
	
	end_current_block ();
	set_unit_part (this_unitp->attrs, EXEC_PART);
}
/*}}}*/

/*{{{  Block *install_cycle_ref (name) -- ok*/
Block *install_cycle_ref (name)
/*---------------------------*/
	char *name;
/*--------------------------------------------------------------*/
/*	Called by 'set_part_and_label ()' when the statement:	*/
/*		CYCLE {name}					*/
/* is encountered.  This function checks whether the CYCLE	*/
/* statement is valid (i.e. in the range of a DO-loop called	*/
/* 'name', or OF any DO-loop if 'name' is NULL).  If it is,	*/
/* the function installs a flag in the DO-loop's 'block' record	*/
/* to signal the presence of the CYCLE instruction, and returns	*/
/* a pointer to that 'block' record.  If not, an error message	*/
/* is given and NULL is returned. 	 			*/
/*								*/
/* N.B. There is no need to install anything in the 'label'	*/
/* record for the DO-loop.  If the CYCLE statement is invalid,	*/
/* an error message is given immediately and no further checks	*/
/* are required when the DO-termination statement is seen.	*/
/*--------------------------------------------------------------*/
{	
	Block  *bp;
	
	/*{{{  check validity of 'CYCLE' statement*/
	if (name) {
		bp = block_name_in_scope (name);
		
		if (!bp) {
			errmsg ("CYCLE statement is not in the range of a DO-loop called %q",
				name);
			return NULL;
		}
	
		switch (bp->type & BL_TYPE_MASK) {
		case TK_DO :
		case TK_DO_WHILE :
		case TK_DO_FOREVER :
			break;	/* ok! */
			
		case TK_IF_CONSTR :
		case TK_ELSE_IF :
		case TK_ELSE :
			errmsg ("CYCLE statement must refer to a DO-loop, but %q is an IF construct",
				name);
			return NULL;
			
		case TK_CASE :
		case TK_DEFAULT :
		case TK_END_SELECT :
			errmsg ("CYCLE statement must refer to a DO-loop, but %q is a CASE construct",
				name);
			return NULL;
			
		}
	}
	else {
		bp = innermost_do_loop ();
	
		if (!bp) {
			errmsg ("CYCLE statement is not within a DO-loop");
			return NULL;
		}
	}
	/*}}}*/

	/*------------------------------------------------------*/
	/* 'CYCLE' reference is valid!  Install a flag in	*/
	/* 'bp->type' to indicate its presence, and return 'bp'.*/
	/*------------------------------------------------------*/	
	bp->type |= BL_CYCLE_REF;

	return  bp;
}
/*}}}*/
/*{{{  Block *install_exit_ref (name) -- ok*/
Block *install_exit_ref (name)
/*--------------------------*/
	char *name;
/*--------------------------------------------------------------*/
/*	Called by 'set_part_and_label ()' when the statement:	*/
/*		EXIT {name}					*/
/* is encountered.  This function checks whether the EXIT	*/
/* statement is valid (i.e. in the range of a DO-loop called	*/
/* 'name', or within any DO-loop if 'name' is NULL).  If it is,	*/
/* the function installs a flag in the DO-loop's 'block' record	*/
/* to signal the presence of the EXIT instruction, and returns	*/
/* a pointer to that 'block' record.  If not, an error message	*/
/* is given and NULL is returned. 	 			*/
/*								*/
/* N.B. There is no need to install anything in the 'label'	*/
/* record for the DO-loop.  If the EXIT statement is invalid,	*/
/* an error message is given immediately and no further checks	*/
/* are required when the DO-termination statement is seen.	*/
/*--------------------------------------------------------------*/
{	
	Block  *bp;
	
	/*{{{  check validity of 'EXIT' statement*/
	if (name) {
		bp = block_name_in_scope (name);
		
		if (!bp) {
			errmsg ("EXIT statement is not in the range of a DO-loop called %q",
				name);
			return NULL;
		}
	
		switch (bp->type & BL_TYPE_MASK) {
		case TK_DO :
		case TK_DO_WHILE :
		case TK_DO_FOREVER :
			break;	/* ok! */
			
		case TK_IF_CONSTR :
		case TK_ELSE_IF :
		case TK_ELSE :
			errmsg ("EXIT statement must refer to a DO-loop, but %q is an IF construct",
				name);
			return NULL;
			
		case TK_CASE :
		case TK_DEFAULT :
		case TK_END_SELECT :
			errmsg ("EXIT statement must refer to a DO-loop, but %q is a CASE construct",
				name);
			return NULL;
		}
	}
	else {
		bp = innermost_do_loop ();
	
		if (!bp) {
			errmsg ("EXIT statement is not within a DO-loop");
			return NULL;
		}
	}
	/*}}}*/

	/*------------------------------------------------------*/
	/* 'EXIT' reference is valid!  Install a flag in	*/
	/* 'bp->type' to indicate its presence, and return 'bp'.*/
	/*------------------------------------------------------*/	
	bp->type |= BL_EXIT_REF;

	return  bp;
}
/*}}}*/

/*--------------------------------------------------------*/
/* (ii) functions to return a pointer to a 'block record' */
/*--------------------------------------------------------*/

/*{{{  Block *lookup_block (int blockno) -- ok*/
Block *lookup_block (blockno)
/*-------------------------*/
	int  blockno;
/*--------------------------------------------------------------*/
/*	This function returns a pointer to block 'blockno' (>0),*/
/* or NULL if the (blockno == 0) or otherwise doesn't exist.	*/
/*--------------------------------------------------------------*/
{
	Blockrec  *blrecp;
	Block  *bp;
	int  n_skiprecs;

	if (blockno < 1 || blockno > n_blocks)
		return  NULL;

	blrecp   = first_blrecp;
	for (n_skiprecs = --blockno / BLOCKS_PER_REC;  n_skiprecs;  --n_skiprecs)
		blrecp = blrecp->next;

	bp = (blrecp->block) + (blockno % BLOCKS_PER_REC);

	return  bp;
}
/*}}}*/
/*{{{  Block *lookup_current_block () -- ok*/
Block *lookup_current_block()
/*-------------------------*/
/*--------------------------------------------------------------*/
/*	This function returns a pointer to the current block,	*/
/* or NULL if there is none.					*/
/*--------------------------------------------------------------*/
{
	return  lookup_block (current_blockno);
}
/*}}}*/

/*-----------------------------------------------------*/
/* (iii) functions to test whether a block is in scope */
/*-----------------------------------------------------*/

/*{{{  bool block_in_scope (int blockno) -- ok*/
bool  block_in_scope (blockno)
/*--------------------------*/
	int  blockno;
/*--------------------------------------------------------------*/
/*	This function returns a boolean result according to	*/
/* whether the block with number 'blockno' is currently 'in	*/
/* scope'.							*/
/*	N.B. There is no guarantee that *any* blocks are in	*/
/* scope when this is called, so test whether 'top_blstkp' is	*/
/* nonzero before dereferencing it!  For example, this function	*/
/* may be called when a 'GOTO 10' stmt is encountered outside	*/
/* of any block (so that 'top_blstkp' == 0), but where label	*/
/* 10 was defined in block no. 1.				*/
/*--------------------------------------------------------------*/
{
	int  *blockp,  *firstp;
	Blockstack  *blstkp;

	
	if (blockno == 0)
		return TRUE;
	else if (n_blocks_in_scope == 0)
		return FALSE;

	/*------------------------------------------------------*/
	/* Now we know that 'blockno' != 0 and that the *are*	*/
	/* some blocks in scope, so we must search the block	*/
	/* stack!						*/
	/*------------------------------------------------------*/
	blockp = current_blockp;
	firstp = top_blstkp->blockno;
	for (  ;  blockp >= firstp;  blockp--)
		if (*blockp == blockno)
			return  TRUE;

	blstkp = top_blstkp->next;	/* may be NULL! */
	for (  ;  blstkp;  blstkp = blstkp->next) {
		firstp = blstkp->blockno;
		blockp = firstp + BLOCKS_PER_STACK_REC - 1;
		for (  ;  blockp >= firstp;  blockp--)
			if (*blockp == blockno)
				return  TRUE;
	}

	return  FALSE;
}
/*}}}*/
/*{{{  Block *block_name_in_scope (char *name) -- ok*/
Block  *block_name_in_scope (name)
/*------------------------------*/
	char  *name;
/*--------------------------------------------------------------*/
/*	If the current line is within a construct called	*/
/* 'name', this function returns a pointer to it, otherwise it	*/
/* returns NULL.						*/
/*	N.B. There is no guarantee that *any* blocks are in	*/
/* scope when this is called, so test whether 'top_blstkp' is	*/
/* nonzero before dereferencing it!  For example, this function	*/
/* may be called when a 'CYCLE name' stmt is encountered	*/
/* outside of any block (so that 'top_blstkp' == 0).		*/
/*--------------------------------------------------------------*/
{
	int  *blockp,  *firstp;
	Blockstack  *blstkp;
	Block  *bp;

	if (n_blocks_in_scope == 0)
		return FALSE;
		
	blockp = current_blockp;
	firstp = top_blstkp->blockno;
	for (  ;  blockp >= firstp;  blockp--) {
		bp = lookup_block (*blockp);
		if (bp->name  &&  str_match (name, bp->name))
			return  bp;
	}

		
	blstkp = top_blstkp->next;	/* may be NULL! */
	for (  ;  blstkp;  blstkp = blstkp->next) {
		firstp = blstkp->blockno;
		blockp = firstp + BLOCKS_PER_STACK_REC - 1;
		for (  ;  blockp >= firstp;  blockp--) {
			bp = lookup_block (*blockp);
			if (bp->name  &&  str_match (name, bp->name))
				return  bp;
		}
	}
		
	return  NULL;
}
/*}}}*/
/*{{{  Block *innermost_do_loop () -- ok*/
Block  *innermost_do_loop ()
/*------------------------*/
/*--------------------------------------------------------------*/
/*	Returns a pointer to the innermost DO-loop that is	*/
/* currently in scope.						*/
/*	N.B. There is no guarantee that *any* blocks are in	*/
/* scope when this is called, so test whether 'top_blstkp' is	*/
/* nonzero before dereferencing it!  For example, this function	*/
/* may be called when a 'CYCLE' stmt is encountered outside of	*/
/* any block (so that 'top_blstkp' == 0).			*/
/*--------------------------------------------------------------*/
{
	int  *blockp,  *firstp;
	Blockstack  *blstkp;
	Block  *bp;

	if (n_blocks_in_scope == 0)
		return FALSE;
		
	blockp = current_blockp;
	firstp = top_blstkp->blockno;
	for (  ;  blockp >= firstp;  blockp--) {
		bp = lookup_block (*blockp);
		if (bp->type & (TK_DO | TK_DO_WHILE | TK_DO_FOREVER))
			return  bp;
	}

		
	blstkp = top_blstkp->next;	/* may be NULL! */
	for (  ;  blstkp;  blstkp = blstkp->next) {
		firstp = blstkp->blockno;
		blockp = firstp + BLOCKS_PER_STACK_REC - 1;
		for (  ;  blockp >= firstp;  blockp--) {
			bp = lookup_block (*blockp);
			if (bp->type & (TK_DO | TK_DO_WHILE | TK_DO_FOREVER))
				return  bp;
		}
	}
		
	return  NULL;
}
/*}}}*/

/*--------------------------------------------------------------*/
/* (iv) function to clear the 'block' records ready for the	*/
/*      next program unit.					*/
/*--------------------------------------------------------------*/

/*{{{  clear_blocks () -- ok*/
clear_blocks ()
/*-----------*/
/*--------------------------------------------------------------*/
/*	Remove all 'block' records, simultaneously checking	*/
/* that all constructs are correctly terminated.		*/
/*--------------------------------------------------------------*/
{
	Block  *bp;
	Blockrec  *blrecp,  *next_blrecp;

	bool  give_errmsg = this_unitp->attrs & PROC_LEVEL;
		/* i.e. not within an INTERFACE block.  We must	*/
		/* be in a proc unit, else this fn isn't called.*/

	
	/*--------------------------------------------------------------*/
	/* (i) If the 'block stack' is not empty, terminate all current	*/
	/* blocks, giving suitable error messages.			*/
	/*	N.B. The following loop calls 'end_current_block ()',	*/
	/* which 'frees' the 'Blockstack' records as they're emptied	*/
	/* and also resets all the global vars associated with them,	*/
	/* namely:							*/
	/*								*/
	/*	n_blocks_in_scope;  current_blockno;  current_blockp;	*/
	/*	top_blstkp;						*/
	/*--------------------------------------------------------------*/
	
	while (n_blocks_in_scope) {
		/*{{{  terminate current block, giving errmsg*/
		
		bp = lookup_current_block ();
		
		switch (bp->type & BL_TYPE_MASK) {
		case TK_DO :
		case TK_DO_WHILE :
		case TK_DO_FOREVER :
			if (give_errmsg) {
				if (bp->do_label)
					errmsg ("\'DO %d\' loop %q beginning on line %d is not terminated",
						bp->do_label, bp->name, bp->constr_lineno);
				else
					errmsg ("DO-loop %q beginning on line %d is not terminated",
						bp->name, bp->constr_lineno);
			}
			end_current_block ();
			break;
		
		case TK_IF_CONSTR :
		case TK_ELSE_IF :
		case TK_ELSE :
			if (give_errmsg)
				errmsg ("IF construct %q beginning on line %d is not terminated",
					bp->name, bp->constr_lineno);
			end_current_block ();
			break;
		
		case TK_CASE :
		case TK_DEFAULT :
			if (give_errmsg)
				errmsg ("CASE construct %q beginning on line %d is not terminated",
					bp->name, bp->constr_lineno);
			end_current_block ();	/* to end the TK_CASE block */
			end_current_block ();	/* to end the TK_END_SELECT block */
			break;
		
		case TK_WHERE_CONSTR :
		case TK_ELSE_WHERE :
			/*------------------------------------------------------*/
			/* These cases shouldn't occur -- if we're in a WHERE	*/
			/* construct, it would have been terminated by		*/
			/* 'set_part_and_label (EXEC_PART, TK_END)'.		*/
			/* (Assuming it's called, that is -- currently the	*/
			/* action for 'end_unit_stmt' doesn't call it!!)	*/
			/*------------------------------------------------------*/
			if (give_errmsg)
				errmsg ("WHERE construct beginning on line %d is not terminated",
					bp->constr_lineno);
			end_current_block ();
			set_unit_part (this_unitp->attrs, EXEC_PART);
			break;
		}
		/*}}}*/
	}
	
	/*--------------------------------------------------------------*/
	/* (ii) Free the 'block' records.				*/
	/* N.B. We don't free 'bp->name', as the names are stored in	*/
	/* 'name buffers' that are separately allocated and freed.	*/
	/*--------------------------------------------------------------*/
	
	for (blrecp = first_blrecp;  blrecp;  blrecp = next_blrecp) {
		next_blrecp = blrecp->next;
		free ((char *) blrecp);
	}
	
	first_blrecp = NULL;
	n_blocks = 0;
}
/*}}}*/

/*}}}*/
/*}}}*/

